from __future__ import annotations as _annotations

import ast
import json
from collections.abc import Iterable, Sequence
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Protocol

from acp.schema import (
    ContentToolCallContent,
    FileEditToolCallContent,
    TerminalToolCallContent,
    TextContentBlock,
    ToolCallLocation,
    ToolCallProgress,
    ToolCallStart,
    ToolCallStatus,
    ToolKind,
)

__all__ = (
    "BrowserProjectionMap",
    "CommandProjectionMap",
    "CommunityFileManagementProjectionMap",
    "CompositeProjectionMap",
    "DeepAgentsProjectionMap",
    "DefaultToolClassifier",
    "FileSystemProjectionMap",
    "FinanceProjectionMap",
    "HttpRequestProjectionMap",
    "ProjectionAwareToolClassifier",
    "ProjectionMap",
    "ToolClassifier",
    "WebFetchProjectionMap",
    "WebSearchProjectionMap",
    "build_tool_progress_update",
    "build_tool_start_update",
    "compose_projection_maps",
    "extract_tool_call_locations",
)

_PATH_KEYS = (
    "destination_path",
    "file_path",
    "filepath",
    "path",
    "source_path",
    "target_path",
)
_OLD_CONTENT_KEYS = ("old_string", "old_text")
_NEW_CONTENT_KEYS = ("content", "new_string", "new_text", "text")
_COMMAND_KEYS = ("command", "cmd", "script", "bash")
_SEARCH_KEYS = ("pattern", "query", "directory", "path")
_OUTPUT_TEXT_KEYS = ("output", "result", "stderr", "stdout", "text")
_TERMINAL_ID_KEYS = ("terminalId", "terminal_id")
_MAX_COMMAND_TITLE_CHARS = 80
_RISKY_COMMAND_PATTERNS = (
    " rm ",
    "chmod ",
    "curl |",
    "dd ",
    " git reset --hard",
    " mkfs",
    " sudo ",
)
_DEFAULT_WEB_SEARCH_TOOL_NAMES = frozenset(
    {
        "brave_search",
        "duckduckgo_results_json",
        "duckduckgo_search",
        "google_search",
        "google_search_results_json",
        "google_serper",
        "google_serper_results_json",
        "jina_search",
        "searchapi",
        "searchapi_results_json",
        "searx_search",
        "searx_search_results",
        "tavily_answer",
        "tavily_search_results_json",
    },
)
_DEFAULT_WEB_FETCH_TOOL_NAMES = frozenset(
    {
        "requests_delete",
        "requests_get",
        "requests_patch",
        "requests_post",
        "requests_put",
    },
)
_DEFAULT_BROWSER_NAVIGATE_TOOL_NAMES = frozenset({"navigate_browser"})
_DEFAULT_BROWSER_READ_TOOL_NAMES = frozenset(
    {"current_webpage", "extract_hyperlinks", "extract_text", "get_elements"},
)
_DEFAULT_BROWSER_ACTION_TOOL_NAMES = frozenset({"click_element", "previous_webpage"})
_DEFAULT_COMMAND_TOOL_NAMES = frozenset({"terminal"})
_DEFAULT_FILE_MANAGEMENT_READ_TOOL_NAMES = frozenset({"read_file"})
_DEFAULT_FILE_MANAGEMENT_WRITE_TOOL_NAMES = frozenset({"write_file"})
_DEFAULT_FILE_MANAGEMENT_SEARCH_TOOL_NAMES = frozenset({"file_search", "list_directory"})
_DEFAULT_FILE_MANAGEMENT_MUTATION_TOOL_NAMES = frozenset({"copy_file", "file_delete", "move_file"})
_DEFAULT_FINANCE_SEARCH_TOOL_NAMES = frozenset({"google_finance", "yahoo_finance_news"})
_DEFAULT_FINANCE_DATASET_TOOL_NAMES = frozenset(
    {"balance_sheets", "cash_flow_statements", "income_statements"},
)
_WEB_SEARCH_QUERY_KEYS = ("query", "q", "search_term", "search_query")
_WEB_FETCH_URL_KEYS = ("url", "href", "link", "uri")
_DIRECTORY_PATH_KEYS = ("dir_path",)
_SEARCH_RESULT_CONTAINER_KEYS = (
    "data",
    "items",
    "organic",
    "organic_results",
    "results",
    "value",
)
_SEARCH_RESULT_TITLE_KEYS = ("title", "name")
_SEARCH_RESULT_URL_KEYS = ("url", "href", "link")
_SEARCH_RESULT_SNIPPET_KEYS = ("snippet", "content", "body", "description", "text")
_FETCH_RESULT_TITLE_KEYS = ("title", "name")
_FETCH_RESULT_CONTENT_KEYS = ("content", "body", "text", "markdown")
_FINANCE_QUERY_KEYS = ("query", "ticker", "symbol")
_FINANCE_PERIOD_KEYS = ("period",)
_MAX_CONTENT_PREVIEW_CHARS = 480
_COMMAND_LIST_KEYS = ("commands",)


def _is_string_keyed_object_dict(value: Any) -> bool:
    return isinstance(value, dict) and all(isinstance(key, str) for key in value)


class ToolClassifier(Protocol):
    def classify(self, tool_name: str, raw_input: Any = None) -> ToolKind: ...

    def approval_policy_key(self, tool_name: str, raw_input: Any = None) -> str: ...


@dataclass(slots=True, frozen=True, kw_only=True)
class ToolProjection:
    content: (
        list[ContentToolCallContent | FileEditToolCallContent | TerminalToolCallContent] | None
    ) = None
    locations: list[ToolCallLocation] | None = None
    status: ToolCallStatus | None = None
    title: str | None = None


class ProjectionMap(Protocol):
    def project_start(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
    ) -> ToolProjection | None: ...

    def project_progress(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
        raw_output: Any = None,
        serialized_output: str,
        status: ToolCallStatus,
    ) -> ToolProjection | None: ...


@dataclass(slots=True, frozen=True, kw_only=True)
class CompositeProjectionMap:
    maps: tuple[ProjectionMap, ...]

    def project_start(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
    ) -> ToolProjection | None:
        return _merge_tool_projections(
            projection_map.project_start(tool_name, cwd=cwd, raw_input=raw_input)
            for projection_map in self.maps
        )

    def project_progress(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
        raw_output: Any = None,
        serialized_output: str = "",
        status: ToolCallStatus = "completed",
    ) -> ToolProjection | None:
        return _merge_tool_projections(
            projection_map.project_progress(
                tool_name,
                cwd=cwd,
                raw_input=raw_input,
                raw_output=raw_output,
                serialized_output=serialized_output,
                status=status,
            )
            for projection_map in self.maps
        )


def compose_projection_maps(
    projection_maps: Sequence[ProjectionMap] | None,
) -> ProjectionMap | None:
    if projection_maps is None:
        return None
    if len(projection_maps) == 0:
        return None
    if len(projection_maps) == 1:
        return projection_maps[0]
    return CompositeProjectionMap(maps=tuple(projection_maps))


@dataclass(slots=True, frozen=True, kw_only=True)
class DefaultToolClassifier:
    def classify(self, tool_name: str, raw_input: Any = None) -> ToolKind:
        del raw_input
        normalized = tool_name.lower()
        if normalized in _DEFAULT_WEB_SEARCH_TOOL_NAMES:
            return "search"
        if normalized in _DEFAULT_WEB_FETCH_TOOL_NAMES:
            return "fetch"
        if normalized in _DEFAULT_BROWSER_NAVIGATE_TOOL_NAMES:
            return "fetch"
        if normalized in _DEFAULT_BROWSER_READ_TOOL_NAMES:
            return "read"
        if normalized in _DEFAULT_COMMAND_TOOL_NAMES:
            return "execute"
        if normalized in _DEFAULT_FILE_MANAGEMENT_READ_TOOL_NAMES:
            return "read"
        if normalized in _DEFAULT_FILE_MANAGEMENT_WRITE_TOOL_NAMES:
            return "edit"
        if normalized in _DEFAULT_FILE_MANAGEMENT_SEARCH_TOOL_NAMES:
            return "search"
        if normalized in _DEFAULT_FILE_MANAGEMENT_MUTATION_TOOL_NAMES:
            return "edit"
        if normalized in _DEFAULT_FINANCE_SEARCH_TOOL_NAMES:
            return "search"
        if normalized in _DEFAULT_FINANCE_DATASET_TOOL_NAMES:
            return "read"
        if normalized.startswith(("fetch_", "download_", "scrape_")):
            return "fetch"
        if normalized in {"glob", "grep", "ls"}:
            return "search"
        if "read" in normalized or normalized.startswith("list_"):
            return "read"
        if any(token in normalized for token in ("write", "edit", "patch", "update", "save")):
            return "edit"
        if any(token in normalized for token in ("bash", "command", "execute", "run", "shell")):
            return "execute"
        if "search" in normalized:
            return "search"
        return "other"

    def approval_policy_key(self, tool_name: str, raw_input: Any = None) -> str:
        del raw_input
        return tool_name


@dataclass(slots=True, frozen=True, kw_only=True)
class FileSystemProjectionMap:
    write_tool_names: frozenset[str] = frozenset()
    read_tool_names: frozenset[str] = frozenset()
    search_tool_names: frozenset[str] = frozenset()
    execute_tool_names: frozenset[str] = frozenset()
    default_search_tool: str | None = None
    search_path_arg: str = "path"
    search_pattern_arg: str = "pattern"
    render_search_results_as_tree: bool = False
    hide_dot_directories_in_tree: bool = False
    tree_root_label: str = "."

    def project_start(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
    ) -> ToolProjection | None:
        del cwd
        if not _is_string_keyed_object_dict(raw_input):
            return None
        if tool_name in self._execute_tool_names():
            command = _command_text(raw_input)
            if command is None:
                return None
            content: list[
                ContentToolCallContent | FileEditToolCallContent | TerminalToolCallContent
            ] = [ContentToolCallContent(type="content", content=_text_block(command))]
            risk_note = _command_risk_note(command)
            if risk_note is not None:
                content.append(
                    ContentToolCallContent(type="content", content=_text_block(risk_note)),
                )
            return ToolProjection(
                content=content,
                title=_format_command_title(command),
            )
        if tool_name in self._write_tool_names():
            path = _first_string(raw_input, _PATH_KEYS)
            new_text = _first_string(raw_input, _NEW_CONTENT_KEYS)
            if path is None or new_text is None:
                return None
            old_text = _first_string(raw_input, _OLD_CONTENT_KEYS) or ""
            return ToolProjection(
                content=[
                    FileEditToolCallContent(
                        type="diff",
                        path=path,
                        old_text=old_text,
                        new_text=new_text,
                    ),
                ],
                locations=[ToolCallLocation(path=path)],
                title=_tool_title(tool_name, path=path),
            )
        if tool_name in self._read_tool_names():
            path = _first_string(raw_input, _PATH_KEYS)
            if path is None:
                return None
            return ToolProjection(
                locations=[ToolCallLocation(path=path)],
                title=_tool_title(tool_name, path=path),
            )
        if tool_name in self._search_tool_names():
            location = _first_string(raw_input, (self.search_path_arg, *_PATH_KEYS))
            search_term = _first_string(raw_input, (self.search_pattern_arg, *_SEARCH_KEYS))
            locations = [ToolCallLocation(path=location)] if location is not None else None
            return ToolProjection(
                locations=locations,
                title=_search_title(tool_name, search_term=search_term, path=location),
            )
        return None

    def project_progress(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
        raw_output: Any = None,
        serialized_output: str,
        status: ToolCallStatus,
    ) -> ToolProjection | None:
        del cwd
        if status != "completed":
            return None
        if tool_name in self._read_tool_names() and _is_string_keyed_object_dict(raw_input):
            path = _first_string(raw_input, _PATH_KEYS)
            if path is None:
                return None
            return ToolProjection(
                content=[
                    FileEditToolCallContent(
                        type="diff",
                        path=path,
                        old_text="",
                        new_text=serialized_output,
                    ),
                ],
                locations=[ToolCallLocation(path=path)],
                title=_tool_title(tool_name, path=path),
            )
        if tool_name in self._execute_tool_names():
            content: list[
                ContentToolCallContent | FileEditToolCallContent | TerminalToolCallContent
            ] = []
            terminal_id = _terminal_id(raw_output)
            if terminal_id is not None:
                content.append(TerminalToolCallContent(type="terminal", terminal_id=terminal_id))
            output_text = _output_text(raw_output, serialized_output)
            if output_text:
                content.append(
                    ContentToolCallContent(type="content", content=_text_block(output_text)),
                )
            return ToolProjection(
                content=content or None,
                title=_command_title_from_input(raw_input),
            )
        if tool_name in self._search_tool_names() and _is_string_keyed_object_dict(raw_input):
            location = _first_string(raw_input, (self.search_path_arg, *_PATH_KEYS))
            search_term = _first_string(raw_input, (self.search_pattern_arg, *_SEARCH_KEYS))
            output_text = _output_text(raw_output, serialized_output)
            if self.render_search_results_as_tree:
                output_text = _render_path_tree(
                    output_text,
                    root_label=location or self.tree_root_label,
                    hide_dot_directories=self.hide_dot_directories_in_tree,
                )
            else:
                output_text = _truncate_text(output_text, limit=_MAX_CONTENT_PREVIEW_CHARS)
            locations = [ToolCallLocation(path=location)] if location is not None else None
            return ToolProjection(
                content=[ContentToolCallContent(type="content", content=_text_block(output_text))],
                locations=locations,
                title=_search_title(tool_name, search_term=search_term, path=location),
            )
        return None

    def _read_tool_names(self) -> frozenset[str]:
        return self.read_tool_names

    def _write_tool_names(self) -> frozenset[str]:
        return self.write_tool_names

    def _search_tool_names(self) -> frozenset[str]:
        if self.default_search_tool is None:
            return self.search_tool_names
        return frozenset((*self.search_tool_names, self.default_search_tool))

    def _execute_tool_names(self) -> frozenset[str]:
        return self.execute_tool_names


@dataclass(slots=True, frozen=True, kw_only=True)
class ProjectionAwareToolClassifier:
    base_classifier: ToolClassifier
    projection_maps: Sequence[ProjectionMap]

    def classify(self, tool_name: str, raw_input: Any = None) -> ToolKind:
        projection_kind = self._projection_kind(tool_name)
        if projection_kind is not None:
            return projection_kind
        return self.base_classifier.classify(tool_name, raw_input)

    def approval_policy_key(self, tool_name: str, raw_input: Any = None) -> str:
        return self.base_classifier.approval_policy_key(tool_name, raw_input)

    def _projection_kind(self, tool_name: str) -> ToolKind | None:
        for projection_map in self.projection_maps:
            if isinstance(projection_map, FileSystemProjectionMap):
                if tool_name in projection_map._read_tool_names():
                    return "read"
                if tool_name in projection_map._write_tool_names():
                    return "edit"
                if tool_name in projection_map._execute_tool_names():
                    return "execute"
                if tool_name in projection_map._search_tool_names():
                    return "search"
            if isinstance(projection_map, CompositeProjectionMap):
                nested_classifier = ProjectionAwareToolClassifier(
                    base_classifier=self.base_classifier,
                    projection_maps=projection_map.maps,
                )
                projection_kind = nested_classifier._projection_kind(tool_name)
                if projection_kind is not None:
                    return projection_kind
        return None


@dataclass(slots=True, frozen=True, kw_only=True)
class WebSearchProjectionMap:
    search_tool_names: frozenset[str] = _DEFAULT_WEB_SEARCH_TOOL_NAMES

    def project_start(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
    ) -> ToolProjection | None:
        del cwd
        if tool_name not in self.search_tool_names:
            return None
        query = _web_search_query(raw_input)
        if query is None:
            return None
        return ToolProjection(
            content=[
                ContentToolCallContent(
                    type="content",
                    content=_text_block(_format_web_search_start(raw_input)),
                ),
            ],
            title=f"Search web for {_single_line_preview(query, limit=_MAX_COMMAND_TITLE_CHARS)}",
        )

    def project_progress(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
        raw_output: Any = None,
        serialized_output: str,
        status: ToolCallStatus,
    ) -> ToolProjection | None:
        del cwd
        if status != "completed" or tool_name not in self.search_tool_names:
            return None
        query = _web_search_query(raw_input)
        return ToolProjection(
            content=[
                ContentToolCallContent(
                    type="content",
                    content=_text_block(_format_web_search_progress(raw_output, serialized_output)),
                ),
            ],
            title=(
                f"Search web for {_single_line_preview(query, limit=_MAX_COMMAND_TITLE_CHARS)}"
                if query is not None
                else None
            ),
        )


@dataclass(slots=True, frozen=True, kw_only=True)
class HttpRequestProjectionMap:
    fetch_tool_names: frozenset[str] = _DEFAULT_WEB_FETCH_TOOL_NAMES

    def project_start(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
    ) -> ToolProjection | None:
        del cwd
        if tool_name not in self.fetch_tool_names:
            return None
        url = _web_fetch_url(raw_input)
        if url is None:
            return None
        method = _http_method_label(tool_name)
        return ToolProjection(
            content=[
                ContentToolCallContent(
                    type="content",
                    content=_text_block(_format_web_fetch_start(raw_input)),
                ),
            ],
            title=f"{method} {_single_line_preview(url, limit=_MAX_COMMAND_TITLE_CHARS)}",
        )

    def project_progress(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
        raw_output: Any = None,
        serialized_output: str,
        status: ToolCallStatus,
    ) -> ToolProjection | None:
        del cwd
        if status != "completed" or tool_name not in self.fetch_tool_names:
            return None
        url = _web_fetch_url(raw_input)
        method = _http_method_label(tool_name)
        return ToolProjection(
            content=[
                ContentToolCallContent(
                    type="content",
                    content=_text_block(_format_web_fetch_progress(raw_output, serialized_output)),
                ),
            ],
            title=(
                f"{method} {_single_line_preview(url, limit=_MAX_COMMAND_TITLE_CHARS)}"
                if url is not None
                else None
            ),
        )


@dataclass(slots=True, frozen=True, kw_only=True)
class WebFetchProjectionMap(HttpRequestProjectionMap):
    """Backward-compatible alias for HTTP request-style projection presets."""


@dataclass(slots=True, frozen=True, kw_only=True)
class BrowserProjectionMap:
    navigate_tool_names: frozenset[str] = _DEFAULT_BROWSER_NAVIGATE_TOOL_NAMES
    read_tool_names: frozenset[str] = _DEFAULT_BROWSER_READ_TOOL_NAMES
    action_tool_names: frozenset[str] = _DEFAULT_BROWSER_ACTION_TOOL_NAMES

    def project_start(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
    ) -> ToolProjection | None:
        del cwd
        if tool_name in self.navigate_tool_names:
            url = _web_fetch_url(raw_input)
            if url is None:
                return None
            return ToolProjection(
                content=[
                    ContentToolCallContent(
                        type="content",
                        content=_text_block(f"URL: {url}"),
                    ),
                ],
                title=f"Navigate {_single_line_preview(url, limit=_MAX_COMMAND_TITLE_CHARS)}",
            )
        if tool_name == "click_element" and _is_string_keyed_object_dict(raw_input):
            selector = _first_string(raw_input, ("selector",))
            if selector is None:
                return None
            return ToolProjection(
                content=[
                    ContentToolCallContent(
                        type="content",
                        content=_text_block(f"Selector: {selector}"),
                    ),
                ],
                title=f"Click {_single_line_preview(selector, limit=_MAX_COMMAND_TITLE_CHARS)}",
            )
        if tool_name == "get_elements" and _is_string_keyed_object_dict(raw_input):
            selector = _first_string(raw_input, ("selector",))
            return ToolProjection(
                content=[
                    ContentToolCallContent(
                        type="content",
                        content=_text_block(
                            (
                                f"Selector: {selector}"
                                if selector is not None
                                else "Inspect page elements."
                            ),
                        ),
                    ),
                ],
                title=(
                    f"Inspect {_single_line_preview(selector, limit=_MAX_COMMAND_TITLE_CHARS)}"
                    if selector is not None
                    else "Inspect page elements"
                ),
            )
        if tool_name in self.read_tool_names:
            return ToolProjection(title=_browser_read_title(tool_name))
        if tool_name in self.action_tool_names:
            return ToolProjection(title=_browser_action_title(tool_name))
        return None

    def project_progress(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
        raw_output: Any = None,
        serialized_output: str,
        status: ToolCallStatus,
    ) -> ToolProjection | None:
        del cwd
        if status != "completed":
            return None
        if tool_name in self.navigate_tool_names:
            return ToolProjection(
                content=[
                    ContentToolCallContent(
                        type="content",
                        content=_text_block(_browser_text_preview(raw_output, serialized_output)),
                    ),
                ],
                title=(
                    f"Navigate {_single_line_preview(url, limit=_MAX_COMMAND_TITLE_CHARS)}"
                    if (url := _web_fetch_url(raw_input)) is not None
                    else None
                ),
            )
        if tool_name == "extract_hyperlinks":
            return ToolProjection(
                content=[
                    ContentToolCallContent(
                        type="content",
                        content=_text_block(
                            _format_browser_link_results(raw_output, serialized_output),
                        ),
                    ),
                ],
                title="Extract hyperlinks",
            )
        if tool_name == "get_elements":
            return ToolProjection(
                content=[
                    ContentToolCallContent(
                        type="content",
                        content=_text_block(
                            _format_browser_element_results(raw_output, serialized_output),
                        ),
                    ),
                ],
                title="Inspect page elements",
            )
        if tool_name in self.read_tool_names or tool_name in self.action_tool_names:
            return ToolProjection(
                content=[
                    ContentToolCallContent(
                        type="content",
                        content=_text_block(_browser_text_preview(raw_output, serialized_output)),
                    ),
                ],
                title=_browser_progress_title(tool_name),
            )
        return None


@dataclass(slots=True, frozen=True, kw_only=True)
class CommandProjectionMap:
    execute_tool_names: frozenset[str] = _DEFAULT_COMMAND_TOOL_NAMES

    def project_start(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
    ) -> ToolProjection | None:
        projection = FileSystemProjectionMap(
            execute_tool_names=self.execute_tool_names,
        ).project_start(
            tool_name,
            cwd=cwd,
            raw_input=raw_input,
        )
        if projection is not None:
            return projection
        if tool_name in self.execute_tool_names:
            return ToolProjection(title="Run shell command")
        return None

    def project_progress(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
        raw_output: Any = None,
        serialized_output: str,
        status: ToolCallStatus,
    ) -> ToolProjection | None:
        projection = FileSystemProjectionMap(
            execute_tool_names=self.execute_tool_names,
        ).project_progress(
            tool_name,
            cwd=cwd,
            raw_input=raw_input,
            raw_output=raw_output,
            serialized_output=serialized_output,
            status=status,
        )
        if projection is not None:
            if projection.title is not None or tool_name not in self.execute_tool_names:
                return projection
            return ToolProjection(
                content=projection.content,
                locations=projection.locations,
                status=projection.status,
                title="Run shell command",
            )
        return None


@dataclass(slots=True, frozen=True, kw_only=True)
class CommunityFileManagementProjectionMap:
    read_tool_names: frozenset[str] = _DEFAULT_FILE_MANAGEMENT_READ_TOOL_NAMES
    write_tool_names: frozenset[str] = _DEFAULT_FILE_MANAGEMENT_WRITE_TOOL_NAMES
    search_tool_names: frozenset[str] = _DEFAULT_FILE_MANAGEMENT_SEARCH_TOOL_NAMES
    mutation_tool_names: frozenset[str] = _DEFAULT_FILE_MANAGEMENT_MUTATION_TOOL_NAMES

    def project_start(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
    ) -> ToolProjection | None:
        projection = FileSystemProjectionMap(
            read_tool_names=self.read_tool_names,
            write_tool_names=self.write_tool_names,
        ).project_start(
            tool_name,
            cwd=cwd,
            raw_input=raw_input,
        )
        if projection is not None:
            return projection
        if tool_name == "file_search" and _is_string_keyed_object_dict(raw_input):
            pattern = _first_string(raw_input, ("pattern",))
            directory = _first_string(raw_input, _DIRECTORY_PATH_KEYS) or "."
            lines = [f"Directory: {directory}"]
            if pattern is not None:
                lines.append(f"Pattern: {pattern}")
            return ToolProjection(
                content=[
                    ContentToolCallContent(type="content", content=_text_block("\n".join(lines))),
                ],
                locations=[ToolCallLocation(path=directory)],
                title=(
                    f"Search `{pattern}` in `{directory}`"
                    if pattern is not None
                    else f"Search files in `{directory}`"
                ),
            )
        if tool_name == "list_directory" and _is_string_keyed_object_dict(raw_input):
            directory = _first_string(raw_input, _DIRECTORY_PATH_KEYS) or "."
            return ToolProjection(
                locations=[ToolCallLocation(path=directory)],
                title=f"List `{directory}`",
            )
        if tool_name in {"copy_file", "move_file"} and _is_string_keyed_object_dict(raw_input):
            source_path = _first_string(raw_input, ("source_path",))
            destination_path = _first_string(raw_input, ("destination_path",))
            if source_path is None or destination_path is None:
                return None
            verb = "Copy" if tool_name == "copy_file" else "Move"
            return ToolProjection(
                content=[
                    ContentToolCallContent(
                        type="content",
                        content=_text_block(f"From: {source_path}\nTo: {destination_path}"),
                    ),
                ],
                locations=[
                    ToolCallLocation(path=source_path),
                    ToolCallLocation(path=destination_path),
                ],
                title=f"{verb} `{source_path}` -> `{destination_path}`",
            )
        if tool_name == "file_delete" and _is_string_keyed_object_dict(raw_input):
            path = _first_string(raw_input, ("file_path",))
            if path is None:
                return None
            return ToolProjection(
                locations=[ToolCallLocation(path=path)],
                title=f"Delete `{path}`",
            )
        return None

    def project_progress(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
        raw_output: Any = None,
        serialized_output: str,
        status: ToolCallStatus,
    ) -> ToolProjection | None:
        projection = FileSystemProjectionMap(
            read_tool_names=self.read_tool_names,
            write_tool_names=self.write_tool_names,
        ).project_progress(
            tool_name,
            cwd=cwd,
            raw_input=raw_input,
            raw_output=raw_output,
            serialized_output=serialized_output,
            status=status,
        )
        if projection is not None:
            return projection
        if status != "completed":
            return None
        if tool_name == "file_search" and _is_string_keyed_object_dict(raw_input):
            directory = _first_string(raw_input, _DIRECTORY_PATH_KEYS) or "."
            pattern = _first_string(raw_input, ("pattern",))
            return ToolProjection(
                content=[
                    ContentToolCallContent(
                        type="content",
                        content=_text_block(
                            _truncate_text(serialized_output, limit=_MAX_CONTENT_PREVIEW_CHARS),
                        ),
                    ),
                ],
                locations=[ToolCallLocation(path=directory)],
                title=(
                    f"Search `{pattern}` in `{directory}`"
                    if pattern is not None
                    else f"Search files in `{directory}`"
                ),
            )
        if tool_name == "list_directory" and _is_string_keyed_object_dict(raw_input):
            directory = _first_string(raw_input, _DIRECTORY_PATH_KEYS) or "."
            return ToolProjection(
                content=[
                    ContentToolCallContent(
                        type="content",
                        content=_text_block(
                            _truncate_text(serialized_output, limit=_MAX_CONTENT_PREVIEW_CHARS),
                        ),
                    ),
                ],
                locations=[ToolCallLocation(path=directory)],
                title=f"List `{directory}`",
            )
        if tool_name in self.mutation_tool_names and serialized_output:
            return ToolProjection(
                title=_file_management_mutation_title(tool_name, raw_input=raw_input),
                content=[
                    ContentToolCallContent(
                        type="content",
                        content=_text_block(
                            _truncate_text(serialized_output, limit=_MAX_CONTENT_PREVIEW_CHARS),
                        ),
                    ),
                ],
                locations=_file_management_locations(raw_input),
            )
        return None


@dataclass(slots=True, frozen=True, kw_only=True)
class FinanceProjectionMap:
    search_tool_names: frozenset[str] = _DEFAULT_FINANCE_SEARCH_TOOL_NAMES
    dataset_tool_names: frozenset[str] = _DEFAULT_FINANCE_DATASET_TOOL_NAMES

    def project_start(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
    ) -> ToolProjection | None:
        del cwd
        if tool_name in self.search_tool_names:
            query = _finance_query(raw_input)
            return ToolProjection(
                content=[
                    ContentToolCallContent(
                        type="content",
                        content=_text_block(
                            (f"Query: {query}" if query is not None else "Lookup financial data."),
                        ),
                    ),
                ],
                title=(f"Search finance for {query}" if query is not None else "Search finance"),
            )
        if tool_name in self.dataset_tool_names:
            title = _finance_dataset_title(tool_name, raw_input=raw_input)
            return ToolProjection(title=title)
        return None

    def project_progress(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
        raw_output: Any = None,
        serialized_output: str,
        status: ToolCallStatus,
    ) -> ToolProjection | None:
        del cwd
        if status != "completed":
            return None
        if tool_name in self.search_tool_names:
            query = _finance_query(raw_input)
            return ToolProjection(
                title=(f"Search finance for {query}" if query is not None else "Search finance"),
                content=[
                    ContentToolCallContent(
                        type="content",
                        content=_text_block(
                            _truncate_text(
                                _output_text(raw_output, serialized_output),
                                limit=_MAX_CONTENT_PREVIEW_CHARS,
                            ),
                        ),
                    ),
                ],
            )
        if tool_name in self.dataset_tool_names:
            return ToolProjection(
                title=_finance_dataset_title(tool_name, raw_input=raw_input),
                content=[
                    ContentToolCallContent(
                        type="content",
                        content=_text_block(
                            _truncate_text(
                                _output_text(raw_output, serialized_output),
                                limit=_MAX_CONTENT_PREVIEW_CHARS,
                            ),
                        ),
                    ),
                ],
            )
        return None


@dataclass(slots=True, frozen=True, kw_only=True)
class DeepAgentsProjectionMap:
    base: FileSystemProjectionMap = field(
        default_factory=lambda: FileSystemProjectionMap(
            read_tool_names=frozenset({"read_file"}),
            write_tool_names=frozenset({"edit_file", "write_file"}),
            search_tool_names=frozenset({"glob", "grep", "ls"}),
            execute_tool_names=frozenset({"execute"}),
        ),
    )

    def project_start(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
    ) -> ToolProjection | None:
        return self.base.project_start(tool_name, cwd=cwd, raw_input=raw_input)

    def project_progress(
        self,
        tool_name: str,
        *,
        cwd: Path | None = None,
        raw_input: Any = None,
        raw_output: Any = None,
        serialized_output: str,
        status: ToolCallStatus,
    ) -> ToolProjection | None:
        return self.base.project_progress(
            tool_name,
            cwd=cwd,
            raw_input=raw_input,
            raw_output=raw_output,
            serialized_output=serialized_output,
            status=status,
        )


def extract_tool_call_locations(raw_input: Any) -> list[ToolCallLocation]:
    if not _is_string_keyed_object_dict(raw_input):
        return []
    path = _first_string(raw_input, _PATH_KEYS)
    if path is None:
        return []
    return [ToolCallLocation(path=path)]


@dataclass(slots=True)
class _PathTreeNode:
    directories: dict[str, _PathTreeNode] = field(default_factory=dict)
    files: set[str] = field(default_factory=set)


def _render_path_tree(
    text: str,
    *,
    root_label: str,
    hide_dot_directories: bool,
) -> str:
    root = _PathTreeNode()
    has_path = False
    truncated = False
    for raw_line in text.splitlines():
        line = raw_line.strip()
        if not line:
            continue
        if line == "...":
            truncated = True
            continue
        normalized = line.removeprefix("./")
        if not normalized:
            continue
        if _path_has_hidden_directory(
            normalized,
            hide_dot_directories=hide_dot_directories,
        ):
            continue
        _insert_tree_path(root, normalized)
        has_path = True
    if not has_path:
        return text
    lines = [f"Tree: {root_label}", ""]
    entries = _render_tree_node(root, prefix="")
    lines.extend(entries)
    if truncated:
        lines.append("...")
    return "\n".join(lines).rstrip()


def _path_has_hidden_directory(path: str, *, hide_dot_directories: bool) -> bool:
    if not hide_dot_directories:
        return False
    parts = [part for part in path.strip("/").split("/") if part]
    directory_parts = parts[:-1] if not path.endswith("/") else parts
    return any(part.startswith(".") for part in directory_parts)


def _insert_tree_path(root: _PathTreeNode, path: str) -> None:
    normalized = path.strip("/")
    if not normalized:
        return
    is_directory = path.endswith("/")
    parts = [part for part in normalized.split("/") if part]
    node = root
    last_index = len(parts) - 1
    for index, part in enumerate(parts):
        is_last = index == last_index
        if is_last and not is_directory:
            node.files.add(part)
            return
        node = node.directories.setdefault(part, _PathTreeNode())


def _render_tree_node(node: _PathTreeNode, *, prefix: str) -> list[str]:
    lines: list[str] = []
    entries: list[tuple[str, str, _PathTreeNode | None]] = []
    for directory_name in sorted(node.directories):
        entries.append(("dir", directory_name, node.directories[directory_name]))
    for file_name in sorted(node.files):
        entries.append(("file", file_name, None))
    for index, (entry_type, name, child) in enumerate(entries):
        is_last = index == len(entries) - 1
        connector = "└── " if is_last else "├── "
        if entry_type == "dir":
            lines.append(f"{prefix}{connector}{name}/")
            child_prefix = f"{prefix}{'    ' if is_last else '│   '}"
            assert child is not None
            lines.extend(_render_tree_node(child, prefix=child_prefix))
            continue
        lines.append(f"{prefix}{connector}{name}")
    return lines


def build_tool_start_update(
    *,
    tool_call_id: str,
    tool_name: str,
    classifier: ToolClassifier,
    raw_input: Any,
    cwd: Path | None,
    projection_map: ProjectionMap | None,
) -> ToolCallStart:
    projected = (
        projection_map.project_start(tool_name, cwd=cwd, raw_input=raw_input)
        if projection_map
        else None
    )
    return ToolCallStart(
        session_update="tool_call",
        tool_call_id=tool_call_id,
        title=(projected.title if projected is not None else None) or tool_name,
        kind=classifier.classify(tool_name, raw_input),
        locations=(
            projected.locations
            if projected is not None and projected.locations is not None
            else None
        )
        or extract_tool_call_locations(raw_input)
        or None,
        raw_input=raw_input,
        status=(projected.status if projected is not None else None) or "in_progress",
        content=projected.content if projected is not None else None,
    )


def build_tool_progress_update(
    *,
    tool_call_id: str,
    tool_name: str,
    classifier: ToolClassifier,
    raw_input: Any,
    raw_output: Any,
    serialized_output: str,
    cwd: Path | None,
    projection_map: ProjectionMap | None,
    status: ToolCallStatus,
) -> ToolCallProgress:
    projected = (
        projection_map.project_progress(
            tool_name,
            cwd=cwd,
            raw_input=raw_input,
            raw_output=raw_output,
            serialized_output=serialized_output,
            status=status,
        )
        if projection_map
        else None
    )
    return ToolCallProgress(
        session_update="tool_call_update",
        tool_call_id=tool_call_id,
        title=projected.title if projected is not None else None,
        kind=classifier.classify(tool_name, raw_input),
        locations=(
            projected.locations
            if projected is not None and projected.locations is not None
            else None
        )
        or extract_tool_call_locations(raw_input)
        or None,
        raw_input=raw_input,
        raw_output=raw_output,
        status=(projected.status if projected is not None else None) or status,
        content=(
            projected.content
            if projected is not None and projected.content is not None
            else (
                [ContentToolCallContent(type="content", content=_text_block(serialized_output))]
                if serialized_output
                else None
            )
        ),
    )


def _merge_tool_projections(
    projections: Iterable[ToolProjection | None],
) -> ToolProjection | None:
    merged_content: list[
        ContentToolCallContent | FileEditToolCallContent | TerminalToolCallContent
    ] = []
    merged_locations: list[ToolCallLocation] = []
    merged_title: str | None = None
    merged_status: ToolCallStatus | None = None
    found = False
    for projection in projections:
        if projection is None:
            continue
        found = True
        if projection.content is not None:
            merged_content.extend(projection.content)
        if projection.locations is not None:
            merged_locations.extend(projection.locations)
        if merged_title is None and projection.title is not None:
            merged_title = projection.title
        if merged_status is None and projection.status is not None:
            merged_status = projection.status
    if not found:
        return None
    return ToolProjection(
        content=merged_content or None,
        locations=merged_locations or None,
        title=merged_title,
        status=merged_status,
    )


def _text_block(text: str) -> TextContentBlock:
    return TextContentBlock(type="text", text=text)


def _first_string(value: dict[str, Any], keys: Sequence[str]) -> str | None:
    for key in keys:
        candidate = value.get(key)
        if isinstance(candidate, str):
            return candidate
    return None


def _format_command_title(command: str) -> str:
    stripped = command.strip()
    if not stripped:
        return "command"
    return _single_line_preview(stripped, limit=_MAX_COMMAND_TITLE_CHARS)


def _single_line_preview(text: str, *, limit: int) -> str:
    normalized = " ".join(text.split())
    if len(normalized) <= limit:
        return normalized
    return f"{normalized[: limit - 1]}…"


def _truncate_text(text: str, *, limit: int) -> str:
    if len(text) <= limit:
        return text
    return f"{text[: limit - 1]}…"


def _http_method_label(tool_name: str) -> str:
    if tool_name.startswith("requests_"):
        return tool_name.removeprefix("requests_").upper()
    return "Fetch"


def _command_text(raw_input: dict[str, Any]) -> str | None:
    command = _first_string(raw_input, _COMMAND_KEYS)
    if command is not None:
        return command
    for key in _COMMAND_LIST_KEYS:
        candidate = raw_input.get(key)
        if isinstance(candidate, str) and candidate.strip():
            return candidate
        if isinstance(candidate, list):
            commands = [
                item.strip() for item in candidate if isinstance(item, str) and item.strip()
            ]
            if commands:
                return "\n".join(commands)
    return None


def _command_risk_note(command: str) -> str | None:
    normalized = f" {command.strip().lower()} "
    if any(pattern in normalized for pattern in _RISKY_COMMAND_PATTERNS):
        return "Potentially risky command"
    return None


def _command_title_from_input(raw_input: Any) -> str | None:
    if not _is_string_keyed_object_dict(raw_input):
        return None
    command = _command_text(raw_input)
    if command is None:
        return None
    return _format_command_title(command)


def _tool_title(tool_name: str, *, path: str | None) -> str | None:
    if path is None:
        return None
    normalized = tool_name.lower()
    if "read" in normalized:
        return f"Read `{path}`"
    if "edit" in normalized or "patch" in normalized or "update" in normalized:
        return f"Edit `{path}`"
    if "write" in normalized or "save" in normalized:
        return f"Write `{path}`"
    return None


def _search_title(tool_name: str, *, search_term: str | None, path: str | None) -> str:
    normalized = tool_name.lower()
    if normalized == "grep" and search_term is not None:
        return f"Search `{search_term}`"
    if normalized == "ls":
        return f"List `{path}`" if path is not None else "List files"
    if normalized == "glob" and search_term is not None:
        return f"Glob `{search_term}`"
    return tool_name


def _output_text(raw_output: Any, serialized_output: str) -> str:
    if _is_string_keyed_object_dict(raw_output):
        for key in _OUTPUT_TEXT_KEYS:
            candidate = raw_output.get(key)
            if isinstance(candidate, str):
                return candidate
    return serialized_output


def _terminal_id(raw_output: Any) -> str | None:
    if not _is_string_keyed_object_dict(raw_output):
        return None
    return _first_string(raw_output, _TERMINAL_ID_KEYS)


@dataclass(slots=True, frozen=True, kw_only=True)
class _NormalizedSearchResult:
    title: str
    url: str | None = None
    snippet: str | None = None


def _parse_structured_value(value: str) -> Any:
    if not value.strip():
        return None
    try:
        return json.loads(value)
    except json.JSONDecodeError:
        pass
    try:
        return ast.literal_eval(value)
    except (SyntaxError, ValueError):
        return None


def _web_search_query(raw_input: Any) -> str | None:
    if isinstance(raw_input, str):
        parsed = _parse_structured_value(raw_input)
        if _is_string_keyed_object_dict(parsed):
            return _first_string(parsed, _WEB_SEARCH_QUERY_KEYS)
        return raw_input.strip() or None
    if not _is_string_keyed_object_dict(raw_input):
        return None
    return _first_string(raw_input, _WEB_SEARCH_QUERY_KEYS)


def _web_fetch_url(raw_input: Any) -> str | None:
    if isinstance(raw_input, str):
        stripped = raw_input.strip()
        if stripped.startswith(("http://", "https://")):
            return stripped
        parsed = _parse_structured_value(stripped)
        if _is_string_keyed_object_dict(parsed):
            return _web_fetch_url(parsed)
        return None
    if not _is_string_keyed_object_dict(raw_input):
        return None
    direct_url = _first_string(raw_input, _WEB_FETCH_URL_KEYS)
    if direct_url is not None:
        return direct_url
    for key in ("data", "input", "payload", "text"):
        nested = raw_input.get(key)
        if isinstance(nested, str):
            parsed = _parse_structured_value(nested)
            if _is_string_keyed_object_dict(parsed):
                return _web_fetch_url(parsed)
    return None


def _format_web_search_start(raw_input: Any) -> str:
    query = _web_search_query(raw_input)
    if query is None:
        return "Searching the web."
    return f"Query: {query}"


def _format_web_fetch_start(raw_input: Any) -> str:
    url = _web_fetch_url(raw_input)
    if url is None:
        return "Fetching web content."
    return f"URL: {url}"


def _format_web_search_progress(raw_output: Any, serialized_output: str) -> str:
    results = _normalized_search_results(raw_output)
    if not results:
        return _truncate_text(serialized_output, limit=_MAX_CONTENT_PREVIEW_CHARS)
    lines: list[str] = []
    for index, result in enumerate(results[:5], start=1):
        lines.append(f"{index}. {result.title}")
        if result.url is not None:
            lines.append(result.url)
        if result.snippet is not None:
            lines.append(_single_line_preview(result.snippet, limit=220))
        lines.append("")
    return "\n".join(lines).rstrip()


def _normalized_search_results(
    raw_output: Any,
) -> tuple[_NormalizedSearchResult, ...] | None:
    parsed_output = raw_output
    if isinstance(raw_output, str):
        parsed_output = _parse_structured_value(raw_output)
    result_rows = _search_result_rows(parsed_output)
    if result_rows is None:
        return None
    normalized: list[_NormalizedSearchResult] = []
    for index, row in enumerate(result_rows, start=1):
        normalized_row = _normalize_search_result_row(row, index=index)
        if normalized_row is not None:
            normalized.append(normalized_row)
    return tuple(normalized) if normalized else None


def _search_result_rows(value: Any) -> list[dict[str, Any]] | None:
    if isinstance(value, list):
        rows = [item for item in value if _is_string_keyed_object_dict(item)]
        return rows or None
    if not _is_string_keyed_object_dict(value):
        return None
    for key in _SEARCH_RESULT_CONTAINER_KEYS:
        nested = value.get(key)
        if isinstance(nested, list):
            rows = [item for item in nested if _is_string_keyed_object_dict(item)]
            if rows:
                return rows
    if any(
        value.get(key) is not None for key in _SEARCH_RESULT_TITLE_KEYS + _SEARCH_RESULT_URL_KEYS
    ):
        return [value]
    return None


def _normalize_search_result_row(
    row: dict[str, Any],
    *,
    index: int,
) -> _NormalizedSearchResult | None:
    title = _first_string(row, _SEARCH_RESULT_TITLE_KEYS)
    url = _first_string(row, _SEARCH_RESULT_URL_KEYS)
    snippet = _first_string(row, _SEARCH_RESULT_SNIPPET_KEYS)
    if title is None and url is None and snippet is None:
        return None
    return _NormalizedSearchResult(
        title=title or f"Result {index}",
        url=url,
        snippet=snippet,
    )


def _format_web_fetch_progress(raw_output: Any, serialized_output: str) -> str:
    parsed_output = raw_output
    if isinstance(raw_output, str):
        parsed_output = _parse_structured_value(raw_output)
    if not _is_string_keyed_object_dict(parsed_output):
        return _truncate_text(serialized_output, limit=_MAX_CONTENT_PREVIEW_CHARS)

    lines: list[str] = []
    url = _first_string(parsed_output, _WEB_FETCH_URL_KEYS)
    if url is not None:
        lines.append(f"URL: {url}")
    title = _first_string(parsed_output, _FETCH_RESULT_TITLE_KEYS)
    if title is not None:
        lines.append(f"Title: {title}")
    content = _first_string(parsed_output, _FETCH_RESULT_CONTENT_KEYS)
    if content is not None:
        lines.extend(("", "Preview:", _truncate_text(content, limit=_MAX_CONTENT_PREVIEW_CHARS)))
    if lines:
        return "\n".join(lines)
    return _truncate_text(serialized_output, limit=_MAX_CONTENT_PREVIEW_CHARS)


def _browser_read_title(tool_name: str) -> str:
    if tool_name == "current_webpage":
        return "Read current page"
    if tool_name == "extract_text":
        return "Extract page text"
    if tool_name == "extract_hyperlinks":
        return "Extract hyperlinks"
    if tool_name == "get_elements":
        return "Inspect page elements"
    return tool_name


def _browser_action_title(tool_name: str) -> str:
    if tool_name == "click_element":
        return "Click element"
    if tool_name == "previous_webpage":
        return "Navigate back"
    return tool_name


def _browser_progress_title(tool_name: str) -> str:
    if tool_name in _DEFAULT_BROWSER_READ_TOOL_NAMES:
        return _browser_read_title(tool_name)
    if tool_name in _DEFAULT_BROWSER_ACTION_TOOL_NAMES:
        return _browser_action_title(tool_name)
    return tool_name


def _file_management_locations(raw_input: Any) -> list[ToolCallLocation] | None:
    if not _is_string_keyed_object_dict(raw_input):
        return None
    locations: list[ToolCallLocation] = []
    for key in ("file_path", "source_path", "destination_path", "dir_path"):
        path = raw_input.get(key)
        if isinstance(path, str):
            locations.append(ToolCallLocation(path=path))
    return locations or None


def _file_management_mutation_title(tool_name: str, *, raw_input: Any) -> str | None:
    if not _is_string_keyed_object_dict(raw_input):
        return None
    if tool_name == "copy_file":
        source_path = _first_string(raw_input, ("source_path",))
        destination_path = _first_string(raw_input, ("destination_path",))
        if source_path is not None and destination_path is not None:
            return f"Copy `{source_path}` -> `{destination_path}`"
    if tool_name == "move_file":
        source_path = _first_string(raw_input, ("source_path",))
        destination_path = _first_string(raw_input, ("destination_path",))
        if source_path is not None and destination_path is not None:
            return f"Move `{source_path}` -> `{destination_path}`"
    if tool_name == "file_delete":
        path = _first_string(raw_input, ("file_path",))
        if path is not None:
            return f"Delete `{path}`"
    return None


def _finance_query(raw_input: Any) -> str | None:
    if isinstance(raw_input, str):
        parsed = _parse_structured_value(raw_input)
        if _is_string_keyed_object_dict(parsed):
            return _first_string(parsed, _FINANCE_QUERY_KEYS)
        return raw_input.strip() or None
    if not _is_string_keyed_object_dict(raw_input):
        return None
    return _first_string(raw_input, _FINANCE_QUERY_KEYS)


def _finance_dataset_title(tool_name: str, *, raw_input: Any) -> str:
    ticker = _finance_query(raw_input)
    period: str | None = None
    if _is_string_keyed_object_dict(raw_input):
        period = _first_string(raw_input, _FINANCE_PERIOD_KEYS)
    label = tool_name.replace("_", " ")
    if label == "balance sheets":
        label = "balance sheets"
    if ticker is None:
        return f"Get {label}"
    if period is None:
        return f"Get {label} for {ticker}"
    return f"Get {label} for {ticker} ({period})"


def _browser_text_preview(raw_output: Any, serialized_output: str) -> str:
    if isinstance(raw_output, str):
        return _truncate_text(raw_output, limit=_MAX_CONTENT_PREVIEW_CHARS)
    return _truncate_text(serialized_output, limit=_MAX_CONTENT_PREVIEW_CHARS)


def _format_browser_link_results(raw_output: Any, serialized_output: str) -> str:
    parsed_output = raw_output
    if isinstance(raw_output, str):
        parsed_output = _parse_structured_value(raw_output)
    if not isinstance(parsed_output, list):
        return _truncate_text(serialized_output, limit=_MAX_CONTENT_PREVIEW_CHARS)
    lines: list[str] = []
    for index, value in enumerate(parsed_output[:10], start=1):
        if isinstance(value, str) and value:
            lines.append(f"{index}. {value}")
    return (
        "\n".join(lines)
        if lines
        else _truncate_text(serialized_output, limit=_MAX_CONTENT_PREVIEW_CHARS)
    )


def _format_browser_element_results(raw_output: Any, serialized_output: str) -> str:
    parsed_output = raw_output
    if isinstance(raw_output, str):
        parsed_output = _parse_structured_value(raw_output)
    if not isinstance(parsed_output, list):
        return _truncate_text(serialized_output, limit=_MAX_CONTENT_PREVIEW_CHARS)
    lines: list[str] = []
    for index, value in enumerate(parsed_output[:10], start=1):
        if not _is_string_keyed_object_dict(value):
            continue
        text = _first_string(value, ("text", "value", "name"))
        selector = _first_string(value, ("selector", "xpath"))
        if text is not None and selector is not None:
            lines.append(f"{index}. {text} ({selector})")
        elif text is not None:
            lines.append(f"{index}. {text}")
        elif selector is not None:
            lines.append(f"{index}. {selector}")
    return (
        "\n".join(lines)
        if lines
        else _truncate_text(serialized_output, limit=_MAX_CONTENT_PREVIEW_CHARS)
    )



from __future__ import annotations as _annotations

import asyncio
import sys
import types
from types import SimpleNamespace
from typing import Any, cast

import pytest
from acp.schema import (
    AvailableCommand,
    SessionConfigOptionBoolean,
    SessionMode,
    SessionModeState,
)
from pydantic_acp.runtime.slash_commands import (
    McpServerInfo,
    _iter_mcp_server_infos,
    _mcp_server_info_from_bridge_metadata,
    _mcp_server_info_from_http_toolset,
    _mcp_server_info_from_mcp_toolset,
    _mcp_server_info_from_session_payload,
    _mcp_server_info_from_stdio_toolset,
    _mcp_target_from_client,
    _toolset_name,
    build_available_commands,
    extract_session_mcp_servers,
    list_agent_mcp_servers,
    list_agent_tools,
    parse_slash_command,
    render_hook_listing,
    render_mcp_server_listing,
    render_mode_message,
    render_model_message,
    render_thinking_message,
    render_tool_listing,
    validate_custom_commands,
    validate_mode_command_ids,
)
from pydantic_ai import ModelRequest, ModelResponse, TextPart
from pydantic_ai.capabilities import Hooks
from pydantic_ai.models.function import AgentInfo, FunctionModel
from pydantic_ai.profiles import ModelProfile
from pydantic_ai.tools import DeferredToolRequests, Tool
from pydantic_ai.toolsets._dynamic import DynamicToolset
from pydantic_ai.toolsets.combined import CombinedToolset
from pydantic_ai.toolsets.wrapper import WrapperToolset
from typing_extensions import Sentinel

from .support import (
    UTC,
    AcpSessionContext,
    AdapterConfig,
    Agent,
    AvailableCommandsUpdate,
    ConfigOptionUpdate,
    CurrentModeUpdate,
    JsonValue,
    MemorySessionStore,
    Path,
    PrepareToolsBridge,
    PrepareToolsMode,
    RecordingClient,
    SlashCommandRequest,
    SlashCommandResult,
    StaticSlashCommand,
    StaticSlashCommandProvider,
    TestModel,
    ThinkingBridge,
    ToolCallProgress,
    ToolCallStart,
    agent_message_texts,
    create_acp_agent,
    datetime,
    text_block,
)

_INVALID_SLASH_VALUE = Sentinel("_INVALID_SLASH_VALUE")


def test_new_session_exposes_current_model_without_explicit_model_selection(
    tmp_path: Path,
) -> None:
    adapter = create_acp_agent(
        agent=Agent(TestModel(model_name="openai:gpt-5-mini", custom_output_text="ok")),
        config=AdapterConfig(session_store=MemorySessionStore()),
    )

    response = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))

    assert response.models is not None
    assert response.models.current_model_id == "openai:gpt-5-mini"
    available_model_ids = [model.model_id for model in response.models.available_models]
    assert "openai:gpt-5-mini" in available_model_ids
    assert "codex:gpt-5.4-mini" in available_model_ids
    assert "codex:gpt-5-mini" not in available_model_ids
    assert response.config_options is not None
    assert response.config_options[0].id == "model"


def test_new_session_emits_available_commands_update(tmp_path: Path) -> None:
    adapter = create_acp_agent(
        agent=Agent(TestModel(model_name="openai:gpt-5-mini", custom_output_text="ok")),
        config=AdapterConfig(session_store=MemorySessionStore()),
    )
    client = RecordingClient()
    adapter.on_connect(client)

    asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))

    command_updates = [
        update for _, update in client.updates if isinstance(update, AvailableCommandsUpdate)
    ]

    assert len(command_updates) == 1
    assert [command.name for command in command_updates[0].available_commands] == [
        "model",
        "tools",
        "hooks",
        "mcp-servers",
    ]


def test_slash_command_render_helpers_cover_empty_states() -> None:
    assert parse_slash_command("hello") is None
    assert parse_slash_command("/") is None
    assert parse_slash_command("/   model") is None
    parsed_unknown = parse_slash_command("/unknown arg")
    assert parsed_unknown is not None
    assert parsed_unknown.name == "unknown"
    assert parsed_unknown.argument == "arg"
    parsed = parse_slash_command(" /MODEL codex:gpt-5.4 ")
    assert parsed is not None
    assert parsed.name == "model"
    assert parsed.argument == "codex:gpt-5.4"
    parsed_mode = parse_slash_command("/PLAN")
    assert parsed_mode is not None
    assert parsed_mode.name == "plan"
    assert parsed_mode.argument is None
    parsed_thinking = parse_slash_command("/THINKING high")
    assert parsed_thinking is not None
    assert parsed_thinking.name == "thinking"
    assert parsed_thinking.argument == "high"
    assert render_mode_message(None) == "Current mode: unavailable"
    assert render_model_message(None) == "Current model: unavailable"
    assert render_thinking_message(None) == "Thinking effort: unavailable"
    assert render_tool_listing([]) == "No tools are currently registered."
    assert (
        render_hook_listing([], projection_map=None)
        == "No Hooks capability callbacks are currently registered."
    )
    assert render_mcp_server_listing([]) == "No MCP servers are currently attached."
    assert (
        render_mcp_server_listing(
            [
                McpServerInfo(
                    name="repo",
                    transport="http",
                    source="session",
                    target="https://repo.example/mcp",
                ),
            ],
        )
        == "MCP servers:\n- repo (http, session): https://repo.example/mcp"
    )


def test_build_available_commands_skips_optional_commands_when_state_is_missing() -> None:
    commands = build_available_commands(
        mode_state=SessionModeState(
            current_mode_id="plan",
            available_modes=[SessionMode(id="plan", name="Plan")],
        ),
        model_state=None,
        config_options=None,
    )

    assert [command.name for command in commands] == [
        "plan",
        "tools",
        "hooks",
        "mcp-servers",
    ]

    thinking_commands = build_available_commands(
        mode_state=None,
        model_state=None,
        config_options=[
            SessionConfigOptionBoolean(
                id="thinking",
                name="Thinking",
                current_value=False,
                type="boolean",
            ),
        ],
    )
    assert [command.name for command in thinking_commands] == [
        "thinking",
        "tools",
        "hooks",
        "mcp-servers",
    ]


def test_validate_mode_command_ids_rejects_invalid_duplicate_and_reserved_values() -> None:
    with pytest.raises(ValueError, match="non-empty"):
        validate_mode_command_ids(["   "])

    with pytest.raises(ValueError, match="cannot contain whitespace"):
        validate_mode_command_ids(["code review"])

    with pytest.raises(ValueError, match="Duplicate ids: review"):
        validate_mode_command_ids(["review", " Review "])

    with pytest.raises(ValueError, match="reserved slash command names"):
        validate_mode_command_ids(["model"])


def test_list_agent_tools_skips_internal_invalid_and_non_tool_entries() -> None:
    agent = Agent(TestModel(custom_output_text="unused"))
    function_toolset = agent._function_toolset
    cast("Any", function_toolset).tools = {
        "acp_hidden": Tool(lambda: "x", takes_ctx=False, name="acp_hidden"),
        1: Tool(lambda: "x", takes_ctx=False, name="bad"),  # type: ignore[dict-item]
        "plain": _INVALID_SLASH_VALUE,
        "visible": Tool(lambda: "x", takes_ctx=False, name="visible"),
    }

    tool_infos = list_agent_tools(agent)

    assert [tool_info.name for tool_info in tool_infos] == ["visible"]
    assert list_agent_tools(cast("Any", SimpleNamespace())) == []


def test_model_slash_command_reports_current_model_and_sets_new_model(
    tmp_path: Path,
) -> None:
    adapter = create_acp_agent(
        agent=Agent(TestModel(model_name="openai:gpt-5-mini", custom_output_text="ok")),
        config=AdapterConfig(session_store=MemorySessionStore()),
    )
    client = RecordingClient()
    adapter.on_connect(client)

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    asyncio.run(adapter.prompt(prompt=[text_block("/model")], session_id=session.session_id))
    asyncio.run(
        adapter.prompt(
            prompt=[text_block("/model openai:gpt-5")],
            session_id=session.session_id,
        ),
    )
    asyncio.run(adapter.prompt(prompt=[text_block("/model")], session_id=session.session_id))

    assert agent_message_texts(client) == [
        "Current model: openai:gpt-5-mini",
        "Current model: openai:gpt-5",
        "Current model: openai:gpt-5",
    ]


def test_mode_slash_commands_switch_modes_and_emit_ui_updates(tmp_path: Path) -> None:
    def passthrough_tools(
        tool_defs: list[Any],
    ) -> list[Any]:
        return list(tool_defs)  # pragma: no cover

    adapter = create_acp_agent(
        agent=Agent(TestModel(model_name="openai:gpt-5-mini", custom_output_text="ok")),
        config=AdapterConfig(
            capability_bridges=[
                PrepareToolsBridge(
                    default_mode_id="ask",
                    modes=[
                        PrepareToolsMode(
                            id="ask",
                            name="Ask",
                            description="Ask mode.",
                            prepare_func=lambda _ctx, tool_defs: passthrough_tools(tool_defs),
                        ),
                        PrepareToolsMode(
                            id="plan",
                            name="Plan",
                            description="Plan mode.",
                            plan_mode=True,
                            prepare_func=lambda _ctx, tool_defs: passthrough_tools(tool_defs),
                        ),
                        PrepareToolsMode(
                            id="agent",
                            name="Agent",
                            description="Agent mode.",
                            plan_tools=True,
                            prepare_func=lambda _ctx, tool_defs: passthrough_tools(tool_defs),
                        ),
                    ],
                ),
            ],
            session_store=MemorySessionStore(),
        ),
    )
    client = RecordingClient()
    adapter.on_connect(client)

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    client.updates.clear()

    asyncio.run(adapter.prompt(prompt=[text_block("/plan")], session_id=session.session_id))
    asyncio.run(adapter.prompt(prompt=[text_block("/agent")], session_id=session.session_id))
    asyncio.run(adapter.prompt(prompt=[text_block("/ask")], session_id=session.session_id))

    assert agent_message_texts(client) == [
        "Current mode: plan",
        "Current mode: agent",
        "Current mode: ask",
    ]
    mode_updates = [update for _, update in client.updates if isinstance(update, CurrentModeUpdate)]
    assert [update.current_mode_id for update in mode_updates] == [
        "plan",
        "agent",
        "ask",
    ]
    config_updates = [
        update for _, update in client.updates if isinstance(update, ConfigOptionUpdate)
    ]
    assert [
        next(option.current_value for option in update.config_options if option.id == "mode")
        for update in config_updates
    ] == [
        "plan",
        "agent",
        "ask",
    ]


def test_available_commands_are_derived_from_configured_modes(tmp_path: Path) -> None:
    adapter = create_acp_agent(
        agent=Agent(TestModel(model_name="openai:gpt-5-mini", custom_output_text="ok")),
        config=AdapterConfig(
            capability_bridges=[
                PrepareToolsBridge(
                    default_mode_id="review",
                    modes=[
                        PrepareToolsMode(
                            id="review",
                            name="Review",
                            description="Review mode.",
                            prepare_func=lambda _ctx, tool_defs: list(tool_defs),
                        ),
                        PrepareToolsMode(
                            id="execute",
                            name="Execute",
                            description="Execution mode.",
                            prepare_func=lambda _ctx, tool_defs: list(tool_defs),
                        ),
                    ],
                ),
                ThinkingBridge(),
            ],
            session_store=MemorySessionStore(),
        ),
    )
    client = RecordingClient()
    adapter.on_connect(client)

    asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))

    command_updates = [
        update for _, update in client.updates if isinstance(update, AvailableCommandsUpdate)
    ]

    assert len(command_updates) == 1
    assert [command.name for command in command_updates[0].available_commands] == [
        "review",
        "execute",
        "model",
        "thinking",
        "tools",
        "hooks",
        "mcp-servers",
    ]


def test_reserved_mode_command_ids_raise_value_error() -> None:
    with pytest.raises(
        ValueError,
        match="reserved slash command names",
    ):
        validate_mode_command_ids(["model"])


def test_thinking_slash_command_updates_ui_state_and_session_config(
    tmp_path: Path,
) -> None:
    observed_thinking: list[Any] = []

    def route_with_thinking(
        messages: list[ModelRequest | ModelResponse],
        info: AgentInfo,
    ) -> ModelResponse:
        del messages
        observed_thinking.append(info.model_request_parameters.thinking)
        return ModelResponse(parts=[TextPart("ok")])

    adapter = create_acp_agent(
        agent=Agent(
            FunctionModel(
                route_with_thinking,
                model_name="thinking-model",
                profile=ModelProfile(supports_thinking=True),
            ),
            output_type=str,
        ),
        config=AdapterConfig(
            capability_bridges=[ThinkingBridge()],
            session_store=MemorySessionStore(),
        ),
    )
    client = RecordingClient()
    adapter.on_connect(client)

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    client.updates.clear()

    asyncio.run(adapter.prompt(prompt=[text_block("/thinking")], session_id=session.session_id))
    asyncio.run(
        adapter.prompt(prompt=[text_block("/thinking HIGH")], session_id=session.session_id),
    )
    asyncio.run(adapter.prompt(prompt=[text_block("/thinking")], session_id=session.session_id))
    asyncio.run(
        adapter.prompt(
            prompt=[text_block("Use the configured thinking effort.")],
            session_id=session.session_id,
        ),
    )

    assert agent_message_texts(client) == [
        "Thinking effort: default",
        "Thinking effort: high",
        "Thinking effort: high",
        "ok",
    ]
    config_updates = [
        update for _, update in client.updates if isinstance(update, ConfigOptionUpdate)
    ]
    thinking_values = [
        next(option.current_value for option in update.config_options if option.id == "thinking")
        for update in config_updates
        if any(option.id == "thinking" for option in update.config_options)
    ]
    assert thinking_values
    assert all(value == "high" for value in thinking_values)
    assert observed_thinking[-1] == "high"


def test_model_slash_command_accepts_codex_models(tmp_path: Path, monkeypatch) -> None:
    session_store = MemorySessionStore()
    adapter = create_acp_agent(
        agent=Agent(TestModel(model_name="openai:gpt-5-mini", custom_output_text="ok")),
        config=AdapterConfig(session_store=session_store),
    )
    client = RecordingClient()
    adapter.on_connect(client)

    fake_module = types.ModuleType("codex_auth_helper")
    fake_model = TestModel(model_name="gpt-5", custom_output_text="codex")

    def create_codex_responses_model(model_id: str, *, instructions: str) -> TestModel:
        assert model_id == "gpt-5"
        assert instructions
        return fake_model

    fake_module.__dict__["create_codex_responses_model"] = create_codex_responses_model
    monkeypatch.setitem(sys.modules, "codex_auth_helper", fake_module)

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    asyncio.run(
        adapter.prompt(
            prompt=[text_block("/model codex:gpt-5")],
            session_id=session.session_id,
        ),
    )
    stored_session = session_store.get(session.session_id)

    assert stored_session is not None
    assert stored_session.session_model_id == "codex:gpt-5"


def test_invalid_selected_model_falls_back_to_default_model(tmp_path: Path) -> None:
    adapter = create_acp_agent(
        agent=Agent(TestModel(model_name="openai:gpt-5-mini", custom_output_text="default")),
        config=AdapterConfig(session_store=MemorySessionStore()),
    )
    client = RecordingClient()
    adapter.on_connect(client)

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    asyncio.run(
        adapter.prompt(
            prompt=[text_block("/model broken:model")],
            session_id=session.session_id,
        ),
    )
    client.updates.clear()

    asyncio.run(
        adapter.prompt(
            prompt=[text_block("Run after invalid model selection.")],
            session_id=session.session_id,
        ),
    )
    asyncio.run(adapter.prompt(prompt=[text_block("/model")], session_id=session.session_id))

    assert agent_message_texts(client) == [
        "default",
        "Current model: openai:gpt-5-mini",
    ]


def test_tools_slash_command_lists_registered_tools(tmp_path: Path) -> None:
    agent = Agent(TestModel(custom_output_text="unused"), output_type=[str, DeferredToolRequests])

    @agent.tool_plain(docstring_format="google")
    def read_repo(path: str) -> str:
        """Read a repository file.

        Args:
            path: Relative path to read.

        """
        return path  # pragma: no cover

    @agent.tool_plain(requires_approval=True)
    def delete_repo(path: str) -> str:
        return path  # pragma: no cover

    adapter = create_acp_agent(
        agent=agent,
        config=AdapterConfig(session_store=MemorySessionStore()),
    )
    client = RecordingClient()
    adapter.on_connect(client)

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    asyncio.run(adapter.prompt(prompt=[text_block("/tools")], session_id=session.session_id))

    assert agent_message_texts(client) == [
        "Available tools:\n- delete_repo [approval]\n- read_repo: Read a repository file.",
    ]


def test_hooks_slash_command_lists_registered_hooks(tmp_path: Path) -> None:
    hooks = Hooks[None]()

    @hooks.on.before_model_request
    async def annotate_request(ctx, request_context):
        del ctx  # pragma: no cover
        return request_context  # pragma: no cover

    @hooks.on.before_tool_execute(tools=["echo"])
    async def audit_echo(ctx, *, call, tool_def, args):
        del ctx, call, tool_def  # pragma: no cover
        return args  # pragma: no cover

    agent = Agent(
        TestModel(custom_output_text="unused"),
        deps_type=type(None),
        capabilities=[hooks],
    )

    @agent.tool_plain
    def echo(text: str) -> str:
        return text  # pragma: no cover

    adapter = create_acp_agent(
        agent=agent,
        config=AdapterConfig(session_store=MemorySessionStore()),
    )
    client = RecordingClient()
    adapter.on_connect(client)

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    asyncio.run(adapter.prompt(prompt=[text_block("/hooks")], session_id=session.session_id))

    assert agent_message_texts(client) == [
        "Registered hooks:\n"
        "- Before Model: annotate_request\n"
        "- Before Tool: audit_echo [tools: echo]",
    ]


def test_mcp_servers_slash_command_extracts_servers_from_agent_toolsets(
    tmp_path: Path,
) -> None:
    expected_agent_infos = [
        McpServerInfo(
            name="remote-sse",
            transport="sse",
            target="https://example.com/sse | prefix=docs",
            source="agent",
        ),
        McpServerInfo(
            name="local-stdio",
            transport="stdio",
            target="python server.py | prefix=fs",
            source="agent",
        ),
        McpServerInfo(
            name="https://example.com/mcp",
            transport="http",
            target="https://example.com/mcp",
            source="agent",
        ),
    ]

    agent = Agent(
        TestModel(custom_output_text="unused"),
    )
    adapter = create_acp_agent(
        agent=agent,
        config=AdapterConfig(session_store=MemorySessionStore()),
    )
    client = RecordingClient()
    adapter.on_connect(client)
    monkeypatch = pytest.MonkeyPatch()
    monkeypatch.setattr(
        "pydantic_acp.runtime.slash_commands.list_agent_mcp_servers",
        lambda candidate_agent: expected_agent_infos if candidate_agent is agent else [],
    )

    try:
        session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
        asyncio.run(
            adapter.prompt(
                prompt=[text_block("/mcp-servers")],
                session_id=session.session_id,
            ),
        )
    finally:
        monkeypatch.undo()

    assert agent_message_texts(client) == [
        "MCP servers:\n"
        "- remote-sse (sse, agent): https://example.com/sse | prefix=docs\n"
        "- local-stdio (stdio, agent): python server.py | prefix=fs\n"
        "- https://example.com/mcp (http, agent): https://example.com/mcp",
    ]


def test_extract_session_mcp_servers_skips_invalid_and_dedupes_sources() -> None:
    session = AcpSessionContext(
        session_id="slash-session",
        cwd=Path("/tmp"),
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
        mcp_servers=[
            {
                "name": "repo-http",
                "transport": "http",
                "url": "https://repo.example/mcp",
            },
            {"transport": "stdio", "command": "python"},
            {
                "name": "repo-stdio",
                "transport": "stdio",
                "command": "python",
                "args": ["a.py"],
            },
        ],
        metadata=cast(
            "dict[str, JsonValue]",
            {
                "pydantic_acp": {
                    "mcp": {
                        "servers": [
                            {
                                "name": "repo-http",
                                "transport": "http",
                                "url": "https://repo.example/mcp",
                            },
                            {
                                "name": "bridge-only",
                                "transport": "sse",
                                "tool_prefix": "docs",
                                "description": "remote docs",
                            },
                            {1: "invalid"},
                        ],
                    },
                },
            },
        ),
    )

    server_infos = extract_session_mcp_servers(session)

    assert [(info.name, info.transport, info.target, info.source) for info in server_infos] == [
        ("repo-http", "http", "https://repo.example/mcp", "session"),
        ("repo-stdio", "stdio", "python a.py", "session"),
        ("bridge-only", "sse", "docs | remote docs", "bridge"),
    ]


def test_extract_session_mcp_servers_covers_agent_dedupes_and_invalid_metadata_shapes() -> None:
    http_cls = type("MCPServerStreamableHTTP", (), {})
    http_cls.__module__ = "pydantic_ai.mcp"
    http_toolset = http_cls()
    cast("Any", http_toolset).url = "https://repo.example/mcp"
    cast("Any", http_toolset).id = "repo-http"

    agent = cast("Any", SimpleNamespace(toolsets=[http_toolset, http_toolset]))
    expected_agent_infos = list_agent_mcp_servers(agent)
    assert len(expected_agent_infos) == 1

    bare_session = AcpSessionContext(
        session_id="slash-invalid-root",
        cwd=Path("/tmp"),
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
        metadata={"pydantic_acp": "bad"},  # type: ignore[dict-item]
    )
    assert extract_session_mcp_servers(bare_session, agent=agent) == expected_agent_infos

    session_bad_mcp = AcpSessionContext(
        session_id="slash-invalid-mcp",
        cwd=Path("/tmp"),
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
        metadata={"pydantic_acp": {"mcp": "bad"}},  # type: ignore[dict-item]
    )
    infos = extract_session_mcp_servers(session_bad_mcp, agent=agent)
    assert len(infos) == 1
    assert infos[0].name == "repo-http"

    session_bad_servers = AcpSessionContext(
        session_id="slash-invalid-servers",
        cwd=Path("/tmp"),
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
        metadata={"pydantic_acp": {"mcp": {"servers": "bad"}}},  # type: ignore[dict-item]
    )
    infos = extract_session_mcp_servers(session_bad_servers, agent=agent)
    assert len(infos) == 1
    assert infos[0].target == "https://repo.example/mcp"

    session_with_duplicates = AcpSessionContext(
        session_id="slash-dupes",
        cwd=Path("/tmp"),
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
        metadata={
            "pydantic_acp": {
                "mcp": {
                    "servers": [
                        {
                            "name": "repo-http",
                            "transport": "http",
                            "url": "https://repo.example/mcp",
                        },
                    ],
                },
            },
        },
    )
    infos = extract_session_mcp_servers(session_with_duplicates, agent=agent)
    assert [(info.name, info.source) for info in infos] == [("repo-http", "agent")]


def test_mcp_helper_parsers_cover_invalid_and_fallback_paths() -> None:
    assert _mcp_server_info_from_session_payload({"transport": "stdio"}) is None
    assert _mcp_server_info_from_session_payload({"name": "demo"}) is None
    stdio_info = _mcp_server_info_from_session_payload(
        {"name": "stdio", "transport": "stdio", "args": ["--serve"]},
    )
    assert stdio_info is not None
    assert stdio_info.target == "<stdio> --serve"
    stdio_command_info = _mcp_server_info_from_session_payload(
        {"name": "stdio", "transport": "stdio", "command": "python", "args": []},
    )
    assert stdio_command_info is not None
    assert stdio_command_info.target == "python"
    http_info = _mcp_server_info_from_session_payload(
        {"name": "http", "transport": "http", "url": "https://demo.test/mcp"},
    )
    assert http_info is not None
    assert http_info.target == "https://demo.test/mcp"

    assert _mcp_server_info_from_bridge_metadata("bad") is None
    assert _mcp_server_info_from_bridge_metadata({"name": "demo"}) is None
    bridge_info = _mcp_server_info_from_bridge_metadata({"name": "demo", "transport": "http"})
    assert bridge_info is not None
    assert bridge_info.target == "<http>"

    stdio_cls = type("MCPServerStdio", (), {})
    stdio_cls.__module__ = "pydantic_ai.mcp"
    stdio_toolset = stdio_cls()
    cast("Any", stdio_toolset).command = ""
    cast("Any", stdio_toolset).args = []
    assert _mcp_server_info_from_stdio_toolset(stdio_toolset) is None
    cast("Any", stdio_toolset).command = "python"
    cast("Any", stdio_toolset).tool_prefix = None
    stdio_result = _mcp_server_info_from_stdio_toolset(stdio_toolset)
    assert stdio_result is not None
    assert stdio_result.target == "python"

    http_cls = type("MCPServerStreamableHTTP", (), {})
    http_cls.__module__ = "pydantic_ai.mcp"
    http_toolset = http_cls()
    cast("Any", http_toolset).url = "https://demo.test/mcp"
    cast("Any", http_toolset).tool_prefix = "repo."
    cast("Any", http_toolset)._id = "toolset-id"
    http_result = _mcp_server_info_from_http_toolset(http_toolset)
    assert http_result is not None
    assert http_result.target == "https://demo.test/mcp | prefix=repo."
    cast("Any", http_toolset).url = ""
    assert _mcp_server_info_from_http_toolset(http_toolset) is None
    assert _toolset_name(SimpleNamespace(), fallback="fallback") == "fallback"


def test_list_agent_mcp_servers_handles_fake_toolsets_and_nested_wrappers() -> None:
    stdio_cls = type("MCPServerStdio", (), {})
    stdio_cls.__module__ = "pydantic_ai.mcp"
    stdio_toolset = stdio_cls()
    cast("Any", stdio_toolset).command = "python"
    cast("Any", stdio_toolset).args = ["server.py"]
    cast("Any", stdio_toolset).tool_prefix = "fs"
    cast("Any", stdio_toolset).id = "local-stdio"

    http_cls = type("MCPServerStreamableHTTP", (), {})
    http_cls.__module__ = "pydantic_ai.mcp"
    http_toolset = http_cls()
    cast("Any", http_toolset).url = "https://example.com/mcp"
    cast("Any", http_toolset).tool_prefix = "docs"
    cast("Any", http_toolset)._id = "remote-http"

    dummy_agent = types.SimpleNamespace(toolsets=[stdio_toolset, http_toolset])
    server_infos = list_agent_mcp_servers(cast("Any", dummy_agent))

    assert [(info.name, info.transport, info.target) for info in server_infos] == [
        ("local-stdio", "stdio", "python server.py | prefix=fs"),
        ("remote-http", "http", "https://example.com/mcp | prefix=docs"),
    ]

    dup_agent = types.SimpleNamespace(toolsets=[stdio_toolset, stdio_toolset])
    deduped = list_agent_mcp_servers(cast("Any", dup_agent))
    assert [(info.name, info.target) for info in deduped] == [
        ("local-stdio", "python server.py | prefix=fs"),
    ]

    assert _iter_mcp_server_infos(CombinedToolset([])) == []
    assert _iter_mcp_server_infos(WrapperToolset(CombinedToolset([]))) == []
    assert _iter_mcp_server_infos(DynamicToolset(lambda _ctx: None)) == []
    combined_infos = _iter_mcp_server_infos(
        CombinedToolset(cast("Any", [http_toolset, stdio_toolset])),
    )
    assert [(info.name, info.transport) for info in combined_infos] == [
        ("remote-http", "http"),
        ("local-stdio", "stdio"),
    ]
    dynamic_toolset = DynamicToolset(lambda _ctx: None)
    dynamic_toolset._toolset = cast("Any", http_toolset)
    dynamic_infos = _iter_mcp_server_infos(dynamic_toolset)
    assert [(info.name, info.transport) for info in dynamic_infos] == [("remote-http", "http")]
    assert _iter_mcp_server_infos(object()) == []


def _make_named_transport(class_name: str, url: str) -> Any:
    transport_cls = type(class_name, (), {})
    transport = transport_cls()
    cast("Any", transport).url = url
    return transport


def test_list_agent_mcp_servers_handles_mcp_toolset() -> None:
    mcp_toolset_cls = type("MCPToolset", (), {})
    mcp_toolset_cls.__module__ = "pydantic_ai.mcp"
    http_toolset = mcp_toolset_cls()
    cast("Any", http_toolset).id = "remote-mcp"
    cast("Any", http_toolset).client = SimpleNamespace(url="https://example.com/mcp")

    http_agent = types.SimpleNamespace(toolsets=[http_toolset])
    http_infos = list_agent_mcp_servers(cast("Any", http_agent))
    assert [(info.name, info.transport, info.target) for info in http_infos] == [
        ("remote-mcp", "http", "https://example.com/mcp"),
    ]

    stdio_toolset = mcp_toolset_cls()
    cast("Any", stdio_toolset)._id = "local-mcp"
    cast("Any", stdio_toolset).client = SimpleNamespace(
        transport=SimpleNamespace(command="python", args=["server.py"]),
    )
    stdio_agent = types.SimpleNamespace(toolsets=[stdio_toolset])
    stdio_infos = list_agent_mcp_servers(cast("Any", stdio_agent))
    assert [(info.name, info.transport, info.target) for info in stdio_infos] == [
        ("local-mcp", "stdio", "python server.py"),
    ]


@pytest.mark.parametrize(
    ("client", "expected_transport", "expected_target"),
    [
        (SimpleNamespace(), "http", "<mcp>"),
        (
            SimpleNamespace(url="https://example.com/mcp"),
            "http",
            "https://example.com/mcp",
        ),
        (
            SimpleNamespace(transport=SimpleNamespace(url="https://example.com/mcp")),
            "http",
            "https://example.com/mcp",
        ),
        (
            SimpleNamespace(
                transport=_make_named_transport(
                    "StreamableHttpSseTransport",
                    "https://example.com/sse",
                ),
            ),
            "sse",
            "https://example.com/sse",
        ),
        (
            SimpleNamespace(
                transport=SimpleNamespace(command="python", args=["server.py", "--serve"]),
            ),
            "stdio",
            "python server.py --serve",
        ),
        (
            SimpleNamespace(transport=SimpleNamespace(command="python", args="not-a-list")),
            "stdio",
            "python",
        ),
        (
            SimpleNamespace(transport=SimpleNamespace(command="python")),
            "stdio",
            "python",
        ),
        (
            SimpleNamespace(
                transport=SimpleNamespace(url="", command=""),
                url="https://fallback.example/mcp",
            ),
            "http",
            "https://fallback.example/mcp",
        ),
    ],
)
def test_mcp_target_from_client_resolves_transport_and_target(
    client: Any,
    expected_transport: str,
    expected_target: str,
) -> None:
    assert _mcp_target_from_client(client) == (expected_transport, expected_target)


def test_mcp_server_info_from_mcp_toolset_returns_none_without_client() -> None:
    mcp_toolset_cls = type("MCPToolset", (), {})
    mcp_toolset_cls.__module__ = "pydantic_ai.mcp"
    toolset = mcp_toolset_cls()

    assert _mcp_server_info_from_mcp_toolset(toolset) is None
    assert _iter_mcp_server_infos(toolset) == []


def test_mcp_server_info_from_mcp_toolset_builds_server_info_from_client() -> None:
    mcp_toolset_cls = type("MCPToolset", (), {})
    mcp_toolset_cls.__module__ = "pydantic_ai.mcp"
    toolset = mcp_toolset_cls()
    cast("Any", toolset).id = "remote-mcp"
    cast("Any", toolset).client = SimpleNamespace(
        transport=_make_named_transport("StreamableHttpSseTransport", "https://example.com/sse"),
    )

    server_info = _mcp_server_info_from_mcp_toolset(toolset)
    assert server_info == McpServerInfo(
        name="remote-mcp",
        transport="sse",
        target="https://example.com/sse",
        source="agent",
    )


def test_mcp_servers_slash_command_renders_attached_servers(tmp_path: Path) -> None:
    session_store = MemorySessionStore()
    adapter = create_acp_agent(
        agent=Agent(TestModel(model_name="openai:gpt-5-mini", custom_output_text="unused")),
        config=AdapterConfig(session_store=session_store),
    )
    client = RecordingClient()
    adapter.on_connect(client)

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    stored_session = session_store.get(session.session_id)
    assert stored_session is not None
    stored_session.mcp_servers = [
        {"name": "repo-http", "transport": "http", "url": "https://repo.example/mcp"},
        {
            "name": "repo-stdio",
            "transport": "stdio",
            "command": "python",
            "args": ["server.py"],
        },
    ]
    session_store.save(stored_session)
    asyncio.run(adapter.prompt(prompt=[text_block("/mcp-servers")], session_id=session.session_id))

    assert agent_message_texts(client) == [
        "MCP servers:\n"
        "- repo-http (http, session): https://repo.example/mcp\n"
        "- repo-stdio (stdio, session): python server.py",
    ]


def test_extract_session_mcp_servers_dedupes_agent_and_session_duplicates(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    duplicate_agent_infos = [
        SimpleNamespace(
            name="docs",
            transport="http",
            target="https://demo.test/mcp",
            source="agent",
        ),
        SimpleNamespace(
            name="docs",
            transport="http",
            target="https://demo.test/mcp",
            source="agent",
        ),
    ]
    monkeypatch.setattr(
        "pydantic_acp.runtime.slash_commands.list_agent_mcp_servers",
        lambda agent: duplicate_agent_infos,
    )
    session = AcpSessionContext(
        session_id="slash-agent-dedupes",
        cwd=Path("/tmp"),
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
        mcp_servers=[
            {"name": "repo", "transport": "http", "url": "https://repo.example/mcp"},
            {"name": "repo", "transport": "http", "url": "https://repo.example/mcp"},
        ],
    )

    infos = extract_session_mcp_servers(session, agent=cast("Any", _INVALID_SLASH_VALUE))

    assert [(info.name, info.source) for info in infos] == [
        ("docs", "agent"),
        ("repo", "session"),
    ]


def test_invalid_selected_model_does_not_leave_failed_tool_updates(
    tmp_path: Path,
) -> None:
    adapter = create_acp_agent(
        agent=Agent(TestModel(model_name="openai:gpt-5-mini", custom_output_text="default")),
        config=AdapterConfig(session_store=MemorySessionStore()),
    )
    client = RecordingClient()
    adapter.on_connect(client)

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    asyncio.run(
        adapter.prompt(
            prompt=[text_block("/model broken:model")],
            session_id=session.session_id,
        ),
    )
    client.updates.clear()

    asyncio.run(
        adapter.prompt(
            prompt=[text_block("Run after invalid model selection.")],
            session_id=session.session_id,
        ),
    )

    tool_failures = [
        update
        for _, update in client.updates
        if isinstance(update, ToolCallProgress) and update.status == "failed"
    ]
    assert tool_failures == []


def test_static_custom_slash_command_is_advertised_and_handled(tmp_path: Path) -> None:
    provider = StaticSlashCommandProvider(
        commands=[
            StaticSlashCommand(
                command=AvailableCommand(name="hello", description="Say hello."),
                handler=lambda request: SlashCommandResult(
                    text=f"hello {request.argument}",
                    updates=[
                        ToolCallStart(
                            session_update="tool_call",
                            tool_call_id="custom-command",
                            title="Custom command",
                            kind="execute",
                            status="completed",
                        ),
                    ],
                ),
            ),
        ],
    )
    adapter = create_acp_agent(
        agent=Agent(TestModel(custom_output_text="model-output")),
        config=AdapterConfig(
            session_store=MemorySessionStore(),
            slash_command_provider=provider,
        ),
    )
    client = RecordingClient()
    adapter.on_connect(client)

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    asyncio.run(adapter.prompt(prompt=[text_block("/hello world")], session_id=session.session_id))
    asyncio.run(adapter.prompt(prompt=[text_block("/missing")], session_id=session.session_id))

    command_update = next(
        update for _, update in client.updates if isinstance(update, AvailableCommandsUpdate)
    )
    assert "hello" in [command.name for command in command_update.available_commands]
    assert any(
        isinstance(update, ToolCallStart) and update.tool_call_id == "custom-command"
        for _, update in client.updates
    )
    assert agent_message_texts(client) == ["hello world", "model-output"]


def test_custom_slash_command_can_skip_text_and_surface_refresh(tmp_path: Path) -> None:
    provider = StaticSlashCommandProvider(
        commands=[
            StaticSlashCommand(
                command=AvailableCommand(name="silent", description="Emit update only."),
                handler=lambda request: SlashCommandResult(
                    text=None,
                    updates=[
                        ToolCallStart(
                            session_update="tool_call",
                            tool_call_id="silent-command",
                            title="Silent command",
                            kind="execute",
                            status="completed",
                        ),
                    ],
                    refresh_session_surface=False,
                ),
            ),
        ],
    )
    adapter = create_acp_agent(
        agent=Agent(TestModel(custom_output_text="model-output")),
        config=AdapterConfig(
            session_store=MemorySessionStore(),
            slash_command_provider=provider,
        ),
    )
    client = RecordingClient()
    adapter.on_connect(client)

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    client.updates.clear()
    asyncio.run(adapter.prompt(prompt=[text_block("/silent")], session_id=session.session_id))

    assert agent_message_texts(client) == []
    assert [
        update for _, update in client.updates if isinstance(update, AvailableCommandsUpdate)
    ] == []
    assert any(
        isinstance(update, ToolCallStart) and update.tool_call_id == "silent-command"
        for _, update in client.updates
    )


def test_custom_slash_command_can_decline_and_fall_through(tmp_path: Path) -> None:
    class Provider:
        def available_commands(self, session, agent):
            del session, agent
            return [AvailableCommand(name="maybe", description="Maybe handle.")]

        def handle_command(self, request: SlashCommandRequest) -> SlashCommandResult | None:
            del request
            return SlashCommandResult(handled=False)

    adapter = create_acp_agent(
        agent=Agent(TestModel(custom_output_text="model-output")),
        config=AdapterConfig(
            session_store=MemorySessionStore(),
            slash_command_provider=Provider(),
        ),
    )
    client = RecordingClient()
    adapter.on_connect(client)

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    asyncio.run(adapter.prompt(prompt=[text_block("/maybe")], session_id=session.session_id))

    assert agent_message_texts(client) == ["model-output"]


def test_custom_slash_command_collisions_are_rejected(tmp_path: Path) -> None:
    provider = StaticSlashCommandProvider(
        commands=[
            StaticSlashCommand(
                command=AvailableCommand(name="tools", description="Collides."),
                handler=lambda request: SlashCommandResult(text=request.name),
            ),
        ],
    )
    adapter = create_acp_agent(
        agent=Agent(TestModel(custom_output_text="unused")),
        config=AdapterConfig(
            session_store=MemorySessionStore(),
            slash_command_provider=provider,
        ),
    )

    with pytest.raises(ValueError, match="reserved slash command"):
        asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))


def test_validate_custom_commands_rejects_invalid_duplicates_and_mode_collisions() -> None:
    with pytest.raises(ValueError, match="must match"):
        validate_custom_commands(
            [AvailableCommand(name="bad name", description="Invalid.")],
            mode_state=None,
        )

    with pytest.raises(ValueError, match="normalized"):
        validate_custom_commands(
            [AvailableCommand(name=" Demo ", description="Not canonical.")],
            mode_state=None,
        )

    with pytest.raises(ValueError, match="unique"):
        validate_custom_commands(
            [
                AvailableCommand(name="demo", description="First."),
                AvailableCommand(name="demo", description="Second."),
            ],
            mode_state=None,
        )

    with pytest.raises(ValueError, match="active mode ids"):
        validate_custom_commands(
            [AvailableCommand(name="plan", description="Collides with mode.")],
            mode_state=SessionModeState(
                current_mode_id="plan",
                available_modes=[SessionMode(id=" plan ", name="Plan")],
            ),
        )

    validate_custom_commands(
        [AvailableCommand(name="review", description="Does not collide.")],
        mode_state=SessionModeState(
            current_mode_id="plan",
            available_modes=[SessionMode(id=" plan ", name="Plan")],
        ),
    )



from __future__ import annotations as _annotations

import asyncio
from contextvars import ContextVar
from dataclasses import dataclass
from types import SimpleNamespace
from typing import Any, cast

import pytest
from pydantic_acp.hook_projection import HookEvent
from pydantic_acp.runtime.hook_introspection import (
    _call_hook_func,
    _HookUpdateEmitter,
    _override_root_capability,
    _root_capability,
    _summarize_error,
    _summarize_result,
    _tool_filters,
    _tool_name,
    _wrap_hook_entry,
    _wrap_hooks,
    list_agent_hooks,
    observe_agent_hooks,
)
from pydantic_ai.capabilities import (
    CapabilityOrdering,
    CombinedCapability,
    Hooks,
    HookTimeoutError,
)
from pydantic_ai.messages import ModelResponse, ToolCallPart
from typing_extensions import Sentinel

from .support import (
    AdapterConfig,
    Agent,
    HookProjectionMap,
    MemorySessionStore,
    RecordingClient,
    TestModel,
    ToolCallProgress,
    ToolCallStart,
    create_acp_agent,
    text_block,
)

_INVALID_HOOK_VALUE = Sentinel("_INVALID_HOOK_VALUE")


@dataclass
class _TestHookEntry:
    func: Any
    timeout: float | None = None
    tools: frozenset[str] | None = None


class _ReadOnlyHookEntry:
    __slots__ = ("_func", "_timeout", "tools")

    def __init__(
        self,
        func: Any,
        timeout: float | None = None,
        tools: frozenset[str] | None = None,
    ) -> None:
        self._func = func
        self._timeout = timeout
        self.tools = tools

    def __copy__(self) -> _ReadOnlyHookEntry:
        return self

    @property
    def func(self) -> Any:
        return self._func

    @property
    def timeout(self) -> float | None:
        return self._timeout


def test_existing_before_model_request_hook_emits_acp_updates(tmp_path) -> None:
    hooks = Hooks[None]()

    @hooks.on.before_model_request
    async def annotate_request(ctx, request_context):
        del ctx
        return request_context

    adapter = create_acp_agent(
        agent=Agent(
            TestModel(custom_output_text="hooked"),
            name="hooked-agent",
            deps_type=type(None),
            capabilities=[hooks],
        ),
        config=AdapterConfig(session_store=MemorySessionStore()),
    )
    client = RecordingClient()
    adapter.on_connect(client)

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    asyncio.run(
        adapter.prompt(
            prompt=[text_block("Trigger the hook.")],
            session_id=session.session_id,
        ),
    )

    start_updates = [
        update
        for _, update in client.updates
        if isinstance(update, ToolCallStart)
        and update.title == "Hook Before Model (annotate_request)"
    ]
    progress_updates = [
        update
        for _, update in client.updates
        if isinstance(update, ToolCallProgress)
        and update.title == "Hook Before Model (annotate_request)"
    ]

    assert len(start_updates) == 1
    assert start_updates[0].raw_input == {
        "event": "before_model_request",
        "hook": "annotate_request",
    }
    assert start_updates[0].kind == "fetch"
    assert len(progress_updates) == 1
    assert progress_updates[0].status == "completed"
    assert isinstance(progress_updates[0].raw_output, str)
    assert progress_updates[0].raw_output.startswith("messages=")


def test_wrap_run_event_stream_hook_requires_async_iterable_and_emits_failed_update() -> None:
    visible_updates: list[Any] = []

    async def visible_write(update: Any) -> None:
        visible_updates.append(update)

    visible_emitter = _HookUpdateEmitter(
        write_update=visible_write,
        projection_map=HookProjectionMap(),
        run_id="visible-stream",
    )

    def invalid_stream_hook(*args: Any, **kwargs: Any) -> str:
        del args, kwargs
        return "not-a-stream"

    wrapped_entry, changed = _wrap_hook_entry(
        "wrap_run_event_stream",
        _TestHookEntry(invalid_stream_hook),
        emitter=visible_emitter,
    )

    assert changed is True

    async def empty_stream() -> Any:
        if False:  # pragma: no cover
            yield None  # pragma: no cover

    with pytest.raises(TypeError, match="async iterable"):

        async def consume() -> None:
            async for _ in cast("Any", wrapped_entry).func(
                cast("Any", None),
                stream=empty_stream(),
            ):
                pass  # pragma: no cover

        asyncio.run(consume())

    assert any(getattr(update, "status", None) == "failed" for update in visible_updates)


def test_multiple_registered_hook_callbacks_each_emit_updates(tmp_path) -> None:
    hooks = Hooks[None]()

    @hooks.on.before_model_request
    async def first_request_hook(ctx, request_context):
        del ctx
        return request_context

    @hooks.on.before_model_request
    async def second_request_hook(ctx, request_context):
        del ctx
        return request_context

    adapter = create_acp_agent(
        agent=Agent(
            TestModel(custom_output_text="hooked"),
            deps_type=type(None),
            capabilities=[hooks],
        ),
        config=AdapterConfig(session_store=MemorySessionStore()),
    )
    client = RecordingClient()
    adapter.on_connect(client)

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    asyncio.run(
        adapter.prompt(
            prompt=[text_block("Trigger the registered hooks.")],
            session_id=session.session_id,
        ),
    )

    start_titles = [
        update.title for _, update in client.updates if isinstance(update, ToolCallStart)
    ]
    progress_titles = [
        update.title for _, update in client.updates if isinstance(update, ToolCallProgress)
    ]

    assert start_titles.count("Hook Before Model (first_request_hook)") == 1
    assert start_titles.count("Hook Before Model (second_request_hook)") == 1
    assert progress_titles.count("Hook Before Model (first_request_hook)") == 1
    assert progress_titles.count("Hook Before Model (second_request_hook)") == 1


def test_tool_filtered_hook_preserves_tool_metadata(tmp_path) -> None:
    hooks = Hooks[None]()

    @hooks.on.before_tool_execute(tools=["echo"])
    async def audit_echo(ctx, *, call, tool_def, args):
        del ctx, call, tool_def
        return args

    agent = Agent(
        TestModel(call_tools=["echo"], custom_output_text="tool-hooked"),
        deps_type=type(None),
        capabilities=[hooks],
    )

    @agent.tool_plain
    def echo(text: str) -> str:
        return text.upper()

    adapter = create_acp_agent(
        agent=agent,
        config=AdapterConfig(session_store=MemorySessionStore()),
    )
    client = RecordingClient()
    adapter.on_connect(client)

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    asyncio.run(
        adapter.prompt(
            prompt=[text_block("Call the echo tool.")],
            session_id=session.session_id,
        ),
    )

    start_updates = [
        update
        for _, update in client.updates
        if isinstance(update, ToolCallStart)
        and update.title == "Hook Before Tool [echo] (audit_echo)"
    ]
    progress_updates = [
        update
        for _, update in client.updates
        if isinstance(update, ToolCallProgress)
        and update.title == "Hook Before Tool [echo] (audit_echo)"
    ]

    assert len(start_updates) == 1
    assert start_updates[0].raw_input == {
        "event": "before_tool_execute",
        "hook": "audit_echo",
        "tools": ["echo"],
        "tool_name": "echo",
    }
    assert start_updates[0].kind == "execute"
    assert len(progress_updates) == 1
    assert progress_updates[0].status == "completed"
    assert progress_updates[0].raw_output == "dict"


def test_custom_hook_projection_map_can_hide_and_relabel_events(tmp_path) -> None:
    hooks = Hooks[None]()

    @hooks.on.before_model_request
    async def annotate_request(ctx, request_context):
        del ctx
        return request_context

    adapter = create_acp_agent(
        agent=Agent(
            TestModel(custom_output_text="hooked"),
            deps_type=type(None),
            capabilities=[hooks],
        ),
        config=AdapterConfig(session_store=MemorySessionStore()),
        projection_maps=[
            HookProjectionMap(
                event_labels={"before_model_request": "Model Hook"},
                show_hook_name_in_title=False,
                include_tool_filters=False,
            ),
        ],
    )
    client = RecordingClient()
    adapter.on_connect(client)

    session = asyncio.run(adapter.new_session(cwd=str(tmp_path), mcp_servers=[]))
    asyncio.run(
        adapter.prompt(
            prompt=[text_block("Trigger the hook.")],
            session_id=session.session_id,
        ),
    )

    start_update = next(update for _, update in client.updates if isinstance(update, ToolCallStart))
    progress_update = next(
        update for _, update in client.updates if isinstance(update, ToolCallProgress)
    )
    assert start_update.title == "Hook Model Hook"
    assert progress_update.title == "Hook Model Hook"
    assert start_update.raw_input == {
        "event": "before_model_request",
        "hook": "annotate_request",
    }


def test_hook_introspection_private_helpers_cover_noops_invalid_entries_and_timeouts() -> None:
    recorded_updates: list[Any] = []

    async def write_update(update: Any) -> None:
        recorded_updates.append(update)  # pragma: no cover

    hidden_emitter = _HookUpdateEmitter(
        write_update=write_update,
        projection_map=HookProjectionMap(hidden_event_ids=frozenset({"hidden"})),
        run_id="run",
    )
    hidden_event = HookEvent(
        event_id="hidden",
        hook_name="hidden_hook",
        tool_name=None,
        tool_filters=(),
    )
    assert asyncio.run(hidden_emitter.emit_start(event=hidden_event)) is None
    assert asyncio.run(hidden_emitter.emit_progress(tool_call_id=None, event=hidden_event)) is None
    assert asyncio.run(hidden_emitter.emit(event=hidden_event)) is None
    assert recorded_updates == []

    plain_agent = Agent(TestModel(custom_output_text="plain"))
    with observe_agent_hooks(
        plain_agent,
        write_update=write_update,
        projection_map=HookProjectionMap(),
    ):
        pass

    invalid_agent = Agent(TestModel(custom_output_text="invalid"))
    cast("Any", invalid_agent)._root_capability = _INVALID_HOOK_VALUE
    cast("Any", invalid_agent)._override_root_capability = _INVALID_HOOK_VALUE
    with observe_agent_hooks(
        invalid_agent,
        write_update=write_update,
        projection_map=HookProjectionMap(),
    ):
        pass
    assert _root_capability(invalid_agent) is None
    assert _override_root_capability(invalid_agent) is None

    hooks = Hooks[None]()

    async def alpha(ctx, request_context):
        del ctx  # pragma: no cover
        return request_context  # pragma: no cover

    async def skipped_alpha(ctx, request_context):
        del ctx  # pragma: no cover
        return request_context  # pragma: no cover

    async def beta(ctx, *, call, tool_def, args):
        del ctx, call, tool_def  # pragma: no cover
        return args  # pragma: no cover

    async def gamma(ctx, *, output_context, output, handler):
        del ctx, output_context  # pragma: no cover
        return await handler(output)  # pragma: no cover

    async def delta(ctx, *, requests):
        del ctx, requests  # pragma: no cover

    skipped = _TestHookEntry(skipped_alpha)
    skipped.func.__module__ = "pydantic_acp.bridges.hooks"
    tool_entry = _TestHookEntry(beta, tools=frozenset({"z", "a"}))
    broken_entry = _TestHookEntry(alpha)
    cast("Any", broken_entry).func = None
    cast("Any", hooks)._registry = {
        1: [],
        "before_model_request": ["bad", broken_entry, skipped, _TestHookEntry(alpha)],
        "before_tool_execute": [tool_entry],
        "handle_deferred_tool_calls": [_TestHookEntry(delta)],
        "wrap_output_validate": [_TestHookEntry(gamma)],
    }

    listed = list_agent_hooks(Agent(TestModel(custom_output_text="hooked"), capabilities=[hooks]))
    assert [(info.event_id, info.hook_name, info.tool_filters) for info in listed] == [
        ("before_model_request", "alpha", ()),
        ("before_tool_execute", "beta", ("a", "z")),
        ("deferred_tool_calls", "delta", ()),
        ("output_validate", "gamma", ()),
    ]
    assert _tool_filters(cast("Any", SimpleNamespace(tools=None))) == ()
    assert _tool_name({"call": _INVALID_HOOK_VALUE}) is None

    class SilentError(Exception):
        def __str__(self) -> str:
            return ""

    assert _summarize_error(SilentError()) == "SilentError"
    assert _summarize_result(None) == "completed"
    assert _summarize_result(3) == "3"
    assert _summarize_result(ToolCallPart("echo", {"text": "ok"})) == "echo"
    assert _summarize_result(ModelResponse(parts=[])) == "parts=0"

    invalid_registry_hooks = Hooks[Any]()
    cast("Any", invalid_registry_hooks)._registry = _INVALID_HOOK_VALUE
    wrapped_hooks, changed = _wrap_hooks(invalid_registry_hooks, emitter=hidden_emitter)
    assert wrapped_hooks is invalid_registry_hooks
    assert changed is False

    bad_key_hooks = Hooks[Any]()
    cast("Any", bad_key_hooks)._registry = {1: []}
    wrapped_hooks, changed = _wrap_hooks(bad_key_hooks, emitter=hidden_emitter)
    assert wrapped_hooks is bad_key_hooks
    assert changed is False

    bad_entry_hooks = Hooks[Any]()
    cast("Any", bad_entry_hooks)._registry = {"before_run": ["bad"]}
    wrapped_hooks, changed = _wrap_hooks(bad_entry_hooks, emitter=hidden_emitter)
    assert wrapped_hooks is bad_entry_hooks
    assert changed is False

    non_callable_entry = _TestHookEntry(alpha)
    cast("Any", non_callable_entry).func = None
    wrapped_entry, changed = _wrap_hook_entry(
        "before_run",
        non_callable_entry,
        emitter=hidden_emitter,
    )
    assert wrapped_entry is non_callable_entry
    assert changed is False

    skipped_entry = _TestHookEntry(alpha)
    skipped_entry.func.__module__ = "pydantic_acp.bridges.hooks"
    wrapped_entry, changed = _wrap_hook_entry(
        "before_run",
        skipped_entry,
        emitter=hidden_emitter,
    )
    assert wrapped_entry is skipped_entry
    assert changed is False

    visible_updates: list[Any] = []

    async def visible_write(update: Any) -> None:
        visible_updates.append(update)

    visible_emitter = _HookUpdateEmitter(
        write_update=visible_write,
        projection_map=HookProjectionMap(),
        run_id="visible",
    )

    async def exploding(*args: Any, **kwargs: Any) -> Any:
        del args, kwargs
        raise RuntimeError("boom")

    wrapped_error_entry, changed = _wrap_hook_entry(
        "before_tool_execute",
        _TestHookEntry(exploding),
        emitter=visible_emitter,
    )
    assert changed is True
    with pytest.raises(RuntimeError, match="boom"):
        asyncio.run(
            cast("Any", wrapped_error_entry).func(
                cast("Any", None),
                call=ToolCallPart("echo", {"text": "ok"}),
                tool_def=cast("Any", _INVALID_HOOK_VALUE),
                args={"text": "ok"},
            ),
        )
    assert any(getattr(update, "status", None) == "failed" for update in visible_updates)

    async def slow() -> None:
        await asyncio.sleep(0.01)

    with pytest.raises(HookTimeoutError):
        asyncio.run(_call_hook_func(slow, timeout=0, hook_name="before_run"))

    real_agent = Agent(
        TestModel(custom_output_text="observed"),
        deps_type=type(None),
        capabilities=[Hooks[None]()],
    )
    override_var = ContextVar("_override_root_capability", default=None)
    cast("Any", real_agent)._override_root_capability = override_var
    cast("Any", real_agent)._root_capability = CombinedCapability(capabilities=[])
    with observe_agent_hooks(
        real_agent,
        write_update=write_update,
        projection_map=HookProjectionMap(),
    ):
        pass


def test_wrap_hooks_preserves_hook_ordering() -> None:
    recorded_updates: list[Any] = []

    async def write_update(update: Any) -> None:
        recorded_updates.append(update)  # pragma: no cover

    emitter = _HookUpdateEmitter(
        write_update=write_update,
        projection_map=HookProjectionMap(),
        run_id="ordered",
    )

    async def alpha(ctx, request_context):
        del ctx  # pragma: no cover
        return request_context  # pragma: no cover

    hooks = Hooks[Any](ordering=CapabilityOrdering(position="outermost"))
    cast("Any", hooks)._registry = {"before_model_request": [_TestHookEntry(alpha)]}

    wrapped_hooks, changed = _wrap_hooks(hooks, emitter=emitter)

    assert changed is True
    assert wrapped_hooks.get_ordering() == hooks.get_ordering()


def test_hook_introspection_handles_missing_registry_and_unreplaceable_entries() -> None:
    recorded_updates: list[Any] = []

    async def write_update(update: Any) -> None:
        recorded_updates.append(update)  # pragma: no cover

    emitter = _HookUpdateEmitter(
        write_update=write_update,
        projection_map=HookProjectionMap(),
        run_id="hook-branches",
    )

    assert (
        asyncio.run(
            emitter.emit_progress(
                tool_call_id="hook:hook-branches:1",
                event=HookEvent(
                    event_id="before_run",
                    hook_name="progress_hidden",
                    tool_name=None,
                    tool_filters=(),
                    status=None,
                ),
            ),
        )
        is None
    )
    assert recorded_updates == []

    invalid_agent = Agent(TestModel(custom_output_text="plain"))
    cast("Any", invalid_agent)._root_capability = _INVALID_HOOK_VALUE
    assert list_agent_hooks(invalid_agent) == []

    hooks = Hooks[Any]()
    cast("Any", hooks)._registry = _INVALID_HOOK_VALUE
    agent = Agent(TestModel(custom_output_text="hooked"), capabilities=[hooks])
    assert list_agent_hooks(agent) == []

    async def alpha(ctx, request_context):
        del ctx  # pragma: no cover
        return request_context  # pragma: no cover

    read_only_entry = _ReadOnlyHookEntry(alpha, timeout=1.0)
    wrapped_entry, changed = _wrap_hook_entry(
        "before_run",
        read_only_entry,
        emitter=emitter,
    )

    assert wrapped_entry is read_only_entry
    assert changed is False
